import json
import threading
import datetime
import uuid
import inspect
from enum import IntEnum
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Set, Optional, Callable, Type
from typing_extensions import Annotated, Self

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator, InstanceOf, field_validator
from pydantic_core import PydanticCustomError

import versionhq as vhq
from versionhq.task.evaluation import Evaluation, EvaluationItem
from versionhq.tool.model import Tool, ToolSet, BaseTool
from versionhq.tool.rag_tool import RagTool
from versionhq.tool.gpt.web_search import GPTToolWebSearch
from versionhq.tool.gpt.file_search import GPTToolFileSearch
from versionhq.tool.gpt.cua import GPTToolCUA
from versionhq._utils import process_config, Logger, UsageMetrics, ErrorType


class TaskExecutionType(IntEnum):
    """
    Enumeration to store task execution types of independent tasks without dependencies.
    """
    SYNC = 1
    ASYNC = 2


class ResponseField(BaseModel):
    """
    A class to store a response format that will generate a JSON schema.
    One layer of nested child is acceptable.
    """

    title: str = Field(default=None, description="title of the field")
    data_type: Type = Field(default=None)
    items: Optional[Type] = Field(default=None, description="store data type of the array items")
    properties: Optional[List[BaseModel]] = Field(default=None, description="store dict items in ResponseField format")
    required: bool = Field(default=True)
    nullable: bool = Field(default=False)
    config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="additional rules")


    @model_validator(mode="after")
    def validate_instance(self) -> Self:
        """
        Validate the model instance based on the given `data_type`. (An array must have `items`, dict must have properties.)
        """

        if self.data_type is list and self.items is None:
            self.items = str

        if self.data_type is dict or (self.data_type is list and self.items is dict):
            if self.properties is None:
                raise PydanticCustomError("missing_properties", "The dict type has to set the properties.", {})

            else:
                for item in self.properties:
                    if not isinstance(item, ResponseField):
                        raise PydanticCustomError("invalid_properties", "Properties field must input in ResponseField format.", {})

        return self


    def _format_props(self) -> Dict[str, Any]:
        """
        Structure valid properties from the ResponseField object. 1 layer of nested child is accepted.
        """
        from versionhq.llm.llm_vars import SchemaType

        schema_type = SchemaType(type=self.data_type).convert()
        props: Dict[str, Any] = {}

        if self.data_type is list:
            if self.items is dict:
                nested_p, nested_r = dict(), list()

                if self.properties:
                    for item in self.properties:
                        nested_p.update(**item._format_props())
                        nested_r.append(item.title)

                props = {
                    "type": schema_type,
                    "items": {
                        "type": SchemaType(type=self.items).convert(),
                        "properties": nested_p,
                        "required": nested_r,
                        "additionalProperties": False
                    }
                }

            elif self.items is list:
                props = {
                    "type": schema_type,
                    "items": { "type": SchemaType(type=self.items).convert(), "items": { "type": SchemaType(type=str).convert() }},
                }

            else:
                props = {
                    "type": schema_type,
                    "items": { "type": SchemaType(type=self.items).convert() },
                }


        elif self.data_type is dict:
            p, r = dict(), list()

            if self.properties:
                for item in self.properties:
                    p.update(**item._format_props())
                    r.append(item.title)

            props = {
                "type": schema_type,
                "properties": p,
                "required": r,
                "additionalProperties": False
            }

        else:
            props = {
                "type": schema_type,
                "nullable": self.nullable,
            }

        return { self.title: { **props, **self.config }} if self.config else { self.title: props }


    def _convert(self, value: Any) -> Any:
        """
        Convert the given value to the ideal data type.
        """
        try:
            if self.type is Any:
                pass
            elif self.type is int:
                return int(value)
            elif self.type is float:
                return float(value)
            elif self.type is list or self.type is dict:
                return json.loads(eval(str(value)))
            elif self.type is str:
                return str(value)
            else:
                return value
        except:
            return value


    def _annotate(self, value: Any) -> Annotated:
        """
        Address Pydantic's `create_model`
        """
        return Annotated[self.type, value] if isinstance(value, self.type) else Annotated[str, str(value)]


    def create_pydantic_model(self, result: Dict, base_model: InstanceOf[BaseModel] | Any) -> Any:
        """
        Create a Pydantic model from the given result.
        """
        for k, v in result.items():
            if k is not self.title:
                pass
            elif type(v) is not self.type:
                v = self._convert(v)
                setattr(base_model, k, v)
            else:
                setattr(base_model, k, v)
        return base_model


class TaskOutput(BaseModel):
    """
    A class to store the final output of the given task in raw (string), json_dict, and pydantic class formats.
    """
    task_id: UUID4 = Field(default_factory=uuid.uuid4)
    raw: str = Field(default="")
    json_dict: Dict[str, Any] = Field(default=None)
    pydantic: Optional[Any] = Field(default=None)
    tool_output: Optional[Any] = Field(default=None, description="stores tool result when the task takes tool output as its final output")
    callback_output: Optional[Any] = Field(default=None, description="stores task or agent callback outcome")
    annotations: Optional[Dict[str, Any]] = Field(default=None)
    evaluation: Optional[InstanceOf[Evaluation]] = Field(default=None, description="stores overall evaluation of the task output. stored in ltm")
    usage: Optional[UsageMetrics] = Field(default=None)


    def _fetch_value_of(self, key: str = None) -> Any:
        """Returns a value to the given key."""

        if not key:
            return None

        if self.pydantic and hasattr(self.pydantic, key):
            return getattr(self.pydantic, key)

        elif self.json_dict and key in self.json_dict:
            return self.json_dict[key]

        else:
            return None


    def _to_context_prompt(self) -> str:
        """Formats prompt context in text formats from the final response."""

        context = ""
        match self.final:
            case dict() | self.pydantic:
                try:
                    context = json.dumps(self.final)
                except:
                    context = str(self.final)
            case _:
                context = str(self.final)

        return context


    def evaluate(self, task) -> Evaluation:
        """
        Evaluate the output based on the criteria, score each from 0 to 1 scale, and raise suggestions for future improvement.
        """
        from versionhq.task.TEMPLATES.Description import EVALUATE, SHOTS

        self.evaluation = Evaluation() if not self.evaluation else self.evaluation

        eval_criteria = task.eval_criteria if task.eval_criteria else  ["accuracy", "completeness", "conciseness", ]
        fsl_prompt = ""

        if task.fsls:
            fsl_prompt = SHOTS.format(c=task.fsls[0], w=task.fsls[1] if len(task.fsls) > 1 else "")
        else:
            fsl_prompt = self.evaluation._draft_fsl_prompt(task_description=task.description)

        for item in eval_criteria:
            description = EVALUATE.format(task_description=task.description, task_output=self.raw, eval_criteria=str(item))
            description = description + fsl_prompt if fsl_prompt else description

            task_eval = Task(description=description, response_schema=EvaluationItem)
            res = task_eval.execute(agent=self.evaluation.eval_by)

            if res.pydantic:
                item = EvaluationItem(
                    score=res.pydantic.score,
                    weight=res.pydantic.weight,
                    suggestion=res.pydantic.suggestion,
                    criteria=res.pydantic.criteria
                )
                self.evaluation.items.append(item)

            else:
                try:
                    item = EvaluationItem(
                        score=float(res.json_dict["score"]),
                        weight=float(res.json_dict["weight"]),
                        suggestion=res.json_dict["suggestion"],
                        criteria=res.json_dict["criteria"]
                    )
                    self.evaluation.items.append(item)
                except Exception as e:
                    Logger(verbose=True).log(level="error", message=f"Failed to convert the evaluation items: {str(e)}", color="red")
                    pass

        return self.evaluation


    @property
    def final(self) -> Any:
        """Returns final output from the task."""

        output = None

        if self.callback_output:
            output = self.callback_output

        elif self.tool_output and str(self.tool_output) == self.raw: # tool_output_as_final
            output = self.tool_output

        else:
            output = self.pydantic if self.pydantic else self.json_dict if self.json_dict else self.raw

        return output


    @property
    def aggregate_score(self) -> float | int:
        return self.evaluation.aggregate_score if self.evaluation is not None else 0


    @property
    def json_string(self) -> Optional[str]:
        return json.dumps(self.json_dict)


    def __str__(self) -> str:
        return str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw


class Task(BaseModel):
    """
    A class that stores independent task information and handles task executions.
    """

    __hash__ = object.__hash__
    _original_description: str = PrivateAttr(default=None)
    config: Optional[Dict[str, Any]] = Field(default=None, description="values to set on Task class")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="unique identifier for the object, not set by user")
    name: Optional[str] = Field(default=None)
    description: str = Field(description="Description of the actual task")
    response_schema: Optional[Type[BaseModel] | List[ResponseField]] = Field(default=None, description="stores response format")
    is_multimodal: bool = False

    # tool usage
    tools: Optional[List[Any]] = Field(default_factory=list, description="tools that the agent can use aside from their tools")
    can_use_agent_tools: bool = Field(default=True, description="whether the agent can use their own tools when executing the task")
    tool_res_as_final: bool = Field(default=False, description="when set True, tools res will be stored in the `TaskOutput`")

    image: Optional[str] = Field(default=None, description="absolute file path or url in string")
    file: Optional[str] = Field(default=None, description="absolute file path or url in string")
    audio: Optional[str] = Field(default=None,  description="absolute file path or url in string")

    # test run
    should_test_run: bool = Field(default=False)
    human: bool = Field(default=False)
    _pfg: Any = None

    # executing
    execution_type: TaskExecutionType = Field(default=TaskExecutionType.SYNC)
    allow_delegation: bool = Field(default=False, description="whether to delegate the task to another agent")
    callback: Optional[Callable] = Field(default=None, description="callback to be executed after the task is completed.")
    callback_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict, description="kwargs for the callback when the callback is callable")

    # evaluation
    should_evaluate: bool = Field(default=False, description="True to run the evaluation flow")
    eval_criteria: Optional[List[str]] = Field(default_factory=list, description="stores a list of criteria to evaluate the outcome")
    fsls: Optional[list[str]] = Field(default=None, description="stores ideal/weak responses")

    # recording
    _delegations: int = 0
    processed_agents: Set[str] = Field(default_factory=set, description="store keys of the agents that executed the task")
    output: Optional[TaskOutput] = Field(default=None, description="store the final TaskOutput object")


    @model_validator(mode="before")
    @classmethod
    def process_config(cls, values: Dict[str, Any]) -> None:
        return process_config(values_to_update=values, model_class=cls)


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_required_fields(self) -> Self:
        required_fields = ["description",]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError( f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_tools(self) -> Self:
        if self.tools:
            tool_list = []
            for item in self.tools:
                match item:
                    case Tool() | ToolSet() | BaseTool() | RagTool() | GPTToolCUA() | GPTToolFileSearch() | GPTToolWebSearch():
                        tool_list.append(item)
                    case dict():
                        tool = None
                        try:
                            tool = Tool(**item)
                        except:
                            try:
                                tool = RagTool(**item)
                            except:
                                pass
                        if tool:
                            tool_list.append(tool)
                    case _:
                        if type(item) == callable:
                            tool_list.append(Tool(func=item))
                        else:
                            pass

            self.tools = tool_list
        return self


    def _structure_response_format(self, data_type: str = "object", model_provider: str = "gemini") -> Dict[str, Any] | None:
        """Structures `response_schema` to the LLM response format."""

        from versionhq.task.structured_response import StructuredOutput

        response_format: Dict[str, Any] = None

        if model_provider == "openrouter":
            return response_format

        else:
            if self.response_schema:
                if isinstance(self.response_schema, list):
                    properties, required_fields = {}, []
                    for i, item in enumerate(self.response_schema):
                        if isinstance(item, ResponseField):
                            properties.update(item._format_props())
                            required_fields.append(item.title)

                    response_schema = {
                        "type": data_type,
                        "properties": properties,
                        "required": required_fields,
                        "additionalProperties": False,
                    }
                    response_format = {
                        "type": "json_schema",
                        "json_schema": { "name": "outcome", "schema": response_schema }
                    }

                elif issubclass(self.response_schema, BaseModel):
                    response_format = StructuredOutput(response_format=self.response_schema, provider=model_provider)._format()

        return response_format


    def _sanitize_raw_output(self, raw: str) -> Dict[str, str]:
        """Sanitizes raw output and prepare for json.loads"""

        import re
        import ast

        output, j = None, None
        r = str(raw).strip()
        r = r.replace("true", "True").replace("false", "False").replace("```json", '"').replace("```", '"').replace('\n', '').replace('\\', '')
        r = re.sub("^'", '"', r)
        r = re.sub(r"'\b", '"', r)
        r = r.strip()
        r = r.replace("  ", "")

        try:
            output = json.loads(r)
        except:
            try: j = json.dumps(eval(r))
            except:
                try: j = json.dumps(str(r))
                except: j = r
            output = json.loads(j)

        if isinstance(output, dict):
            return output["json_schema"] if "json_schema" in output else output
        else:
            try:
                output = ast.literal_eval(j)
            except:
                output = ast.literal_eval(r)

            return output["json_schema"] if isinstance(output, dict) and "json_schema" in output else output if isinstance(output, dict) else { "output": str(r) }


    def _create_json_output(self, raw: str) -> Dict[str, Any]:
        """Creates JSON output from the raw output."""

        output = None

        if raw is None or raw == "":
            Logger().log(level="warning", message="The model returned an empty response. Returning an empty dict.", color="yellow")
            output = { "output": "" }
            return output

        try:
            output = json.loads(raw)
            if isinstance(output, dict):
                return output["json_schema"] if "json_schema" in output else output
            else:
                output = self._sanitize_raw_output(raw=raw)
                return output
        except:
            output = self._sanitize_raw_output(raw=raw)
            return output


    def _create_pydantic_output(self, raw: str = None, json_dict: Dict[str, Any] = None) -> InstanceOf[BaseModel] | None:
        """
        Create pydantic output from raw or json_dict output.
        """

        if self.response_schema and not isinstance(self.response_schema, list):
            output_pydantic = self.response_schema
            try:
                json_dict = json_dict if json_dict else self._create_json_output(raw=raw)

                for k, v in json_dict.items():
                    setattr(output_pydantic, k, v)
            except:
                pass
            return output_pydantic

        else:
            return None


    def interpolate_inputs(self, inputs: Dict[str, Any]) -> None:
        """
        Interpolate inputs into the task description.
        """
        self._original_description = self.description

        if inputs:
            self.description = self._original_description.format(**inputs)


    def _create_short_and_long_term_memories(self, agent: Any, task_output: TaskOutput) -> None:
        """
        After the task execution, create and save short-term/long-term memories in the storage.
        """
        from versionhq.agent.model import Agent
        from versionhq.memory.model import ShortTermMemory, MemoryMetadata, LongTermMemory

        agent = agent if isinstance(agent, Agent) else Agent(role=str(agent), goal=str(agent), with_memory=True)

        if agent.with_memory == False:
            return None

        try:
            evaluation = task_output.evaluation if task_output.evaluation else None
            memory_metadata = evaluation._create_memory_metadata() if evaluation else MemoryMetadata()

            agent.short_term_memory = agent.short_term_memory if agent.short_term_memory else ShortTermMemory(agent=agent, embedder_config=agent.embedder_config)
            agent.short_term_memory.save(
                task_description=str(self.description),
                task_output=str(task_output.raw),
                agent=str(agent.role),
                metadata=memory_metadata
            )

            agent.long_term_memory = agent.long_term_memory if agent.long_term_memory else LongTermMemory()
            agent.long_term_memory.save(
                task_description=str(self.description),
                task_output=str(task_output.raw),
                agent=str(agent.role),
                metadata=memory_metadata
            )

        except AttributeError as e:
            Logger().log(level="error", message=f"Missing attributes for long term memory: {str(e)}", color="red")
            pass

        except Exception as e:
            Logger().log(level="error", message=f"Failed to add to the memory: {str(e)}", color="red")
            pass


    def _build_agent_from_task(self, task_description: str = None) -> InstanceOf["vhq.Agent"]:
        task_description = task_description if task_description else self.description
        if not task_description:
            Logger().log(level="error", message="Task is missing the description.", color="red")
            pass

        agent = vhq.Agent(goal=task_description, role=task_description, maxit=1) #! REFINEME
        return agent


    def _select_agent_to_delegate(self, agent: Any = None) -> Any | None: # return agent object or None
        """
        Creates or selects an agent to delegate the given task and returns Agent object else None.
        """

        from versionhq.agent.model import Agent

        if not self.allow_delegation:
            return None

        agent_to_delegate: InstanceOf[Agent] = None

        if not agent:
            agent_to_delegate = self._build_agent_from_task()

        elif agent and not agent.networks:
            agent_to_delegate = Agent(role="vhq-Delegated-Agent", goal=agent.goal, llm=agent.llm)

        else:
            _managers = []
            _members = []
            for network in agent.networks:
                _managers.extend(member.agent for member in network.members if member.is_manager)
                _members.extend(member.agent for member in network.members if not member.is_manager)

            agent_to_delegate = _managers[0] if _managers else _members[0] if _members else Agent(role="vhq-Delegated-Agent", goal=agent.goal, llm=agent.llm)

        return agent_to_delegate


    def _store_logs(self, inputs: Optional[Dict[str, Any]] = {}) -> None:
        from versionhq.storage.task_output_storage import TaskOutputStorageHandler

        TaskOutputStorageHandler().update(task=self, inputs=inputs)


    # task execution
    def execute(self, type: TaskExecutionType = None, agent: "vhq.Agent" = None, context: Any = None) -> TaskOutput | Future[TaskOutput]:
        """A main method to handle task execution."""

        type = type if type else  self.execution_type if self.execution_type else TaskExecutionType.SYNC
        agent = agent if agent else self._build_agent_from_task(task_description=self.description)
        res = None

        if (self.should_test_run or agent.self_learn) and not self._pfg:
            res = self._test_time_computation(agent=agent, context=context)
            return res

        match type:
            case TaskExecutionType.SYNC:
                res = self._execute_sync(agent=agent, context=context)

            case TaskExecutionType.ASYNC:
                res = self._execute_async(agent=agent, context=context)

        return res


    def _execute_sync(self, agent, context: Optional[Any] = None) -> TaskOutput:
        """Executes the task synchronously."""
        return self._execute_core(agent, context)


    def _execute_async(self, agent, context: Optional[Any] = None) -> Future[TaskOutput]:
        """Executes the task asynchronously."""
        future: Future[TaskOutput] = Future()

        def _handle_task_async(self, agent, context: Optional[str], future: Future[TaskOutput]) -> None:
            result = self._execute_core(agent, context)
            future.set_result(result)

        threading.Thread(daemon=True, target=_handle_task_async, args=(agent, context, future)).start()
        return future


    def _execute_core(self, agent, context: Optional[Any]) -> TaskOutput:
        """A core method to execute a single task."""

        start_dt = datetime.datetime.now()
        task_output: InstanceOf[TaskOutput] = None
        user_prompt, dev_prompt = None, None

        if self.allow_delegation == True:
            agent_to_delegate = self._select_agent_to_delegate(agent=agent)
            agent = agent_to_delegate
            self._delegations += 1

        user_prompt, dev_prompt, raw_output, usage = agent.execute_task(task=self, context=context)
        match raw_output:
            case TaskOutput():
                raw_output.task_id = self.id
                raw_output.usage = usage
                task_output = raw_output

            case str():
                json_dict_output = self._create_json_output(raw=raw_output)
                if "outcome" in json_dict_output:
                    json_dict_output = self._create_json_output(raw=str(json_dict_output["outcome"]))

                pydantic_output = self._create_pydantic_output(raw=raw_output, json_dict=json_dict_output)
                task_output = TaskOutput(
                    task_id=self.id,
                    raw=raw_output if raw_output is not None else "",
                    pydantic=pydantic_output,
                    json_dict=json_dict_output,
                    tool_output=raw_output if self.tool_res_as_final else None,
                    usage=usage
                )

            case None | "":
                task_output = TaskOutput(task_id=self.id, raw="", usage=usage)
                task_output.usage.record_errors(type=ErrorType.FORMAT)

            case _:
                task_output = TaskOutput(
                    task_id=self.id,
                    raw=raw_output,
                    tool_output=raw_output if self.tool_res_as_final else None,
                    usage=usage
                )

        self.processed_agents.add(agent.key)

        # if self.output_file: ## disabled for now
        #     content = (
        #         json_output
        #         if json_output
        #         else pydantic_output.model_dump_json() if pydantic_output else result
        #     )
        #     self._save_file(content)

        if self._pfg:
            index = self._pfg.index
            self._pfg.user_prompts.update({ index: user_prompt })
            self._pfg.dev_prompts.update({ index: dev_prompt })

        if task_output.raw:
            if self.should_evaluate:
                task_output.evaluate(task=self)

            self._create_short_and_long_term_memories(agent=agent, task_output=task_output)

            if self.callback and isinstance(self.callback, Callable):
                kwargs = { **self.callback_kwargs, **task_output.json_dict }
                sig = inspect.signature(self.callback)
                valid_keys = [param.name for param in sig.parameters.values() if param.kind == param.POSITIONAL_OR_KEYWORD]
                valid_kwargs = { k: kwargs[k] if  k in kwargs else None for k in valid_keys }
                callback_res = self.callback(**valid_kwargs, **task_output.json_dict )
                task_output.callback_output = callback_res

        end_dt = datetime.datetime.now()
        task_output.usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        if task_output.json_dict and "output" in task_output.json_dict:
            task_output.usage.record_errors(type=ErrorType.FORMAT)

        self.output = task_output
        self._store_logs()
        return task_output


    def _test_time_computation(self, agent, context: Optional[Any]) -> TaskOutput | None:
        """Handles test-time computation."""

        from versionhq.task_graph.model import ReformTriggerEvent
        from versionhq._prompt.model import Prompt
        from versionhq._prompt.auto_feedback import PromptFeedbackGraph

        prompt = Prompt(task=self, agent=agent, context=context)
        pfg = PromptFeedbackGraph(prompt=prompt, should_reform=self.human, reform_trigger_event=ReformTriggerEvent.USER_INPUT if self.human else None)
        pfg = pfg.set_up_graph()
        self._pfg = pfg

        try:
            if self._pfg and self.output is None:
                res, all_outputs = self._pfg.activate()
                if all_outputs:
                    res.usage = self._pfg.usage
                return res
        except:
            self._pfg.usage.record_errors(type=ErrorType.API)
            Logger().log(level="error", message="Failed to execute the task.", color="red")
            return None


    @property
    def key(self) -> str:
        output_format = "json" if self.response_schema else "raw"
        source = [self.description, output_format]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def summary(self) -> str:
        return f"""
Task ID: {str(self.id)}
"Description": {self.description}
"Tools": {", ".join([tool.name for tool in self.tools])}
        """



from versionhq import Product, ProductProvider
from versionhq.tool.composio.params import ComposioAppName


def test_create_product_provider():
    provider = ProductProvider(name="demo", region="US", data_pipelines=["hubspot", "demo crm"] )

    assert provider.id is not None
    assert provider.name == "demo"
    assert isinstance(provider.data_pipelines[0], ComposioAppName)
    assert isinstance(provider.data_pipelines[1], str)
    assert provider.destination_services is None


def test_create_product():
    provider = ProductProvider(name="demo", region="US", data_pipelines=["hubspot", "demo crm"], destination_services=["test"])
    product = Product(name="demo", description="demo", provider=provider, audience="demo", usp="demo", landing_url="www.com", cohort_timeframe=30)

    assert product.id is not None
    assert product.name == "demo"
    assert product.provider.id is not None



def test_doc_core_taskoutput_a():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    def dummy_tool():
        return "dummy"

    def summarize_response(message: str, **kwargs) -> str | None:
        test1 = kwargs["test1"] if kwargs and "test1" in kwargs else None
        test2 = kwargs["test2"] if kwargs and "test2" in kwargs else None
        if test1 and test2:
            return f"""{message}: {test1}, {", ".join(str(test2))}"""
        else:
            return None

    task = vhq.Task(
        description="Research a topic to teach a kid aged 6 about math.",
        response_schema=CustomOutput,
        tools=[dummy_tool],
        callback=summarize_response,
        callback_kwargs=dict(message="Hi! Here is the result: "),
        should_evaluate=True, # triggers evaluation
        eval_criteria=["Uniquness", "Fit to audience",],

    )
    res = task.execute()

    assert res.task_id == task.id
    assert res.raw
    assert res.json_dict
    assert res.pydantic.test1 and res.pydantic.test2
    if res.callback_output:
        assert "Hi! Here is the result: " in res.callback_output
    assert res.tool_output is None
    assert res.evaluation and isinstance(res.evaluation, vhq.Evaluation)



import base64
import datetime
import time
import platform
from typing import List, Dict, Any, Tuple, Literal, get_args

from versionhq._utils import convert_img_url
from versionhq.tool.gpt import openai_client
from versionhq.tool.gpt._enum import GPTSizeEnum
from versionhq._utils import is_valid_enum, UsageMetrics, ErrorType, Logger, is_valid_url, handle_directory

BROWSER = Literal['chromium', 'firefox']
TYPE = Literal["computer_call_output", "computer_use_preview"]
ENV =  Literal["browser", "mac", "windows", "ubuntu"]

class CUAToolSchema:
    type: TYPE = "computer_use_preview"
    environment: ENV = "browser"
    display_width: int = 1024
    display_height: int = 768

    def __init__(self, type: str = None, display_width: int = None, display_height: int = None, environment: str = None):
        self.display_height = display_height if display_height else self.display_height
        self.display_width = display_width if display_width else self.display_width

        if type and type in get_args(TYPE):
            self.type = type

        if environment and environment in get_args(ENV):
            self.environment = environment

    @property
    def schema(self) -> Dict[str, Any]:
        return {
            "type": self.type if isinstance(self.type, str) else self.type.value,
            "display_width": self.display_width,
            "display_height": self.display_height,
            "environment": self.environment if isinstance(self.environment, str) else self.environment.value,
        }


class GPTToolCUA:
    model: str = "computer-use-preview"
    tools: List[CUAToolSchema] = list()
    user_prompt: str = None
    img_url: str = None
    web_url: str = "https://www.google.com"
    browser: BROWSER = "firefox"
    reasoning_effort: str = GPTSizeEnum.MEDIUM.value
    truncation: str = "auto"

    _schema: Dict[str, Any] = dict()
    _response_ids: List[str] = list()
    _call_ids: List[str] = list()
    _calls: Dict[str, Dict[str, Any]] = dict() # stores response_id and raw output object.
    _usage: UsageMetrics = UsageMetrics()
    _logger: Logger = Logger(info_file_save=True, filename="cua-task-{}".format(str(datetime.datetime.now().timestamp())) + ".png")


    def __init__(
        self,
        user_prompt: str,
        tools: List[CUAToolSchema] | CUAToolSchema = None,
        img_url: str = None,
        web_url: str = "https://www.google.com",
        browser: str = "chromium",
        reasoning_effort: GPTSizeEnum | str = None,
        truncation: str = None,
        _usage: UsageMetrics = UsageMetrics()
    ):
        self.user_prompt = user_prompt
        self.web_url = web_url if is_valid_url(web_url) else None
        self.browser = browser if browser in get_args(BROWSER) else 'chromium'
        self.truncation = truncation if truncation else self.truncation
        self._usage = _usage
        self._response_ids = list()
        self._call_ids = list()

        if img_url:
            img_url = convert_img_url(img_url)
            self.img_url = img_url

        if reasoning_effort and is_valid_enum(enum=GPTSizeEnum, val=reasoning_effort):
            self.reasoning_effort = reasoning_effort.value if isinstance(reasoning_effort, GPTSizeEnum) else reasoning_effort

        if tools:
            match tools:
                case list():
                    if self.tools:
                        self.tools.extend(tools)
                    else:
                        self.tools = tools
                case CUAToolSchema():
                    if self.tools:
                        self.tools.append(tools)
                    else:
                        self.tools = [tools]
                case _:
                    pass


    def _structure_schema(self, screenshot: str = None) -> None:
        """Formats args schema for CUA calling."""

        tool_schema = [item.schema for item in self.tools]
        schema = dict()
        inputs = list()
        previous_response_id = self._response_ids[-1] if self._response_ids else None
        # (self._response_ids[-1].startswith("rs") or  self._response_ids[-1].startswith("resp")) else None

        if self._call_ids:
            inputs = [
                {
                    "call_id": self._call_ids[-1],
                    "type": "computer_call_output",
                }
            ]
            if screenshot:
                inputs[0].update({ "output": { "type": "computer_screenshot", "image_url": f"data:image/png;base64,{str(screenshot)}"}})

            # if self._calls:
            #     call = self._calls[self._call_ids[-1]]
            #     if call and call.call_id not in inputs[0]:
            #         inputs.append(call)

            if previous_response_id:
                schema = dict(
                    model=self.model,
                    previous_response_id=previous_response_id,
                    tools=tool_schema,
                    input=inputs,
                    truncation=self.truncation
                )
            else:
                schema = dict(
                    model=self.model,
                    tools=tool_schema,
                    input=inputs,
                    truncation=self.truncation
                )

        else:
            input = [{ "role": "user", "content": self.user_prompt } ]
            img_url = convert_img_url(self.img_url) if self.img_url else None
            if img_url:
                input.append({"type": "input_image", "image_url": f"data:image/png;base64,{img_url}"})

            schema = dict(
                model=self.model,
                tools=tool_schema,
                input=input,
                reasoning={ "effort": self.reasoning_effort},
                truncation=self.truncation
            )

        self._schema = schema
        # return self._schema


    def _run(self, screenshot: str = None) -> Tuple[Dict[str, Any], None, UsageMetrics]:
        raw_res = dict()
        usage = self._usage if self._usage else UsageMetrics()
        start_dt = datetime.datetime.now()

        try:
            self._structure_schema(screenshot=screenshot)
            res = openai_client.responses.create(**self._schema)
            if not res:
                usage.record_errors(ErrorType.TOOL)
            else:
                self._response_ids.append(res.id)
                for item in res.output:

                    match item.type:
                        case "reasoning":
                            reasoning = item.summary[0].text if item.summary and isinstance(item.summary, list) else str(item.summary) if item.summary else ""
                            raw_res.update(dict(reasoning=reasoning))
                            # self._response_ids.append(item.id)

                        case "computer_call":
                            raw_res.update(dict(action=item.action))
                            # self._response_ids.append(item.id)
                            call_id = item.call_id
                            self._call_ids.append(call_id)
                            self._calls.update({ call_id: item })
                        case _:
                            pass
            usage.record_token_usage(**res.usage.__dict__)

        except Exception as e:
            self._logger.log(message=f"Failed to run: {str(e)}", color="red", level="error")
            usage.record_errors(ErrorType.TOOL)

        end_dt = datetime.datetime.now()
        usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        return raw_res, None, usage


    def invoke_playwright(self) -> Dict[str, Any]:
        """Handles computer use loop. Ref. OpenAI official website."""
        try:
            from playwright.sync_api import sync_playwright
        except Exception as e:
            self._logger.log(level="error", message=f"Install Playwright by adding `versionhq[tools]` to requirements.txt or run `uv add playwright`. {str(e)}", color="red")
            raise e

        import os
        os.environ["DEBUG"] = "pw:browser"
        self._logger.log(message="Start computer use.", level="info", color="blue")
        start_dt = datetime.datetime.now()
        res = None

        # try:
        p = sync_playwright().start()
        b = p.firefox if self.browser == "firefox" else p.chromium
        browser = b.launch(headless=True)
        page = browser.new_page()
        if not browser or not page:
            return None, None, None

        if self.web_url:
            page.goto(self.web_url, timeout=3000000, wait_until="load", referer=None)
            time.sleep(3)

        res, _, usage = self._run()
        self._usage.aggregate(metrics=usage)
        actions = [v for k, v in res.items() if k =="action"] if res else []
        action = actions[0] if actions else None

        if action:
            while True:
                x = action.x if hasattr(action, 'x') else 0
                y = action.y if hasattr(action, 'y') else 0
                scroll_x = action.scroll_x if hasattr(action, 'scroll_x') else 0
                scroll_y = action.scroll_y if hasattr(action, 'scroll_y') else 0
                text = action.text if hasattr(action, 'text') else ''
                screenshot_base64 = None
                path = handle_directory(directory_name='_screenshots', filename=f'cua_playwright', ext='png')

                match action.type:
                    case "click":
                        self._logger.log(message="Action: click", color="blue", level="info")
                        button = action.button if hasattr(action, 'button') and (action.button == 'left' or action.button == 'right') else 'left'
                        page.mouse.move(x, y)
                        page.mouse.click(x, y, button=button)
                        time.sleep(1)

                    case "scroll":
                        self._logger.log(message="Action: scroll", color="blue", level="info")
                        page.mouse.move(x, y)
                        page.evaluate(f"window.scrollBy({scroll_x}, {scroll_y})")
                        time.sleep(1)

                    case "move":
                        self._logger.log(message="Action: move", color="blue", level="info")
                        page.mouse.move(x, y)
                        page.evaluate(f"window.scrollBy({scroll_x}, {scroll_y})")
                        time.sleep(1)

                    case "keypress":
                        self._logger.log(message="Action: keypress", color="blue", level="info")
                        keys = action.keys
                        for k in keys:
                            match k.lower():
                                case "enter": page.keyboard.press("Enter")
                                case "space": page.keyboard.press(" ")
                                case _: page.keyboard.press(k)
                        time.sleep(1)

                    case "type":
                        self._logger.log(message="Action: type", color="blue", level="info")
                        page.keyboard.type(text)
                        time.sleep(1)

                    case "wait":
                        self._logger.log(message="Action: wait", color="blue", level="info")
                        time.sleep(3)

                    case "screenshot":
                        self._logger.log(message="Action: screenshot", color="blue", level="info")
                        screenshot_bytes = page.screenshot(path=path)
                        screenshot_base64 = base64.b64encode(screenshot_bytes).decode("utf-8")
                        time.sleep(1)

                    case _:
                        self._logger.log(message=f"Unrecognized action: {action}", level="warning", color="yellow")
                        return False

                if not screenshot_base64:
                    screenshot_bytes = page.screenshot(path=path)
                    screenshot_base64 = base64.b64encode(screenshot_bytes).decode("utf-8")
                    time.sleep(1)

                res, _, usage = self._run(screenshot=screenshot_base64)
                self._usage.aggregate(metrics=usage)
                if not res:
                    usage.record_errors(type=ErrorType.API)
                    break

                actions = [v for k, v in res.items() if k =="action"] if res else []
                action = actions[0] if actions else None
                if not action:
                    break
        else:
            self._usage.record_errors(type=ErrorType.TOOL)

        # except Exception as e:
        #     self._logger.log(message=f"Failed to execute. {str(e)}", color="red", level="error")
        #     browser.close()

        end_dt = datetime.datetime.now()
        self._usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        return res


    def invoke_selenium(self, **kwargs) -> Dict[str, Any]:
        try:
            from selenium import webdriver
            from selenium.webdriver.common.keys import Keys
            from selenium.webdriver.common.action_chains import ActionChains
            from selenium.webdriver.common.actions.action_builder import ActionBuilder
        except Exception as e:
            self._logger.log(level="error", message=f"Install Selenium by `uv pip install versionhq[tools]` or `uv add selenium`. {str(e)}", color="red")
            raise e

        self._logger.log(message="Start computer use", level="info", color="blue")

        start_dt = datetime.datetime.now()

        driver = webdriver.Chrome(options=kwargs) if kwargs else webdriver.Chrome()
        if self.tools:
            driver.set_window_size(height=self.tools[0].display_height, width=self.tools[0].display_width)

        if self.web_url:
            driver.get(self.web_url)
            time.sleep(3)

        res, _, usage = self._run()
        self._logger.log(message=f"Initial response: {res}", color="blue", level="info")
        self._usage.aggregate(metrics=usage)
        actions = [v for k, v in res.items() if k =="action"] if res else []
        action = actions[0] if actions else None
        action_chains = ActionChains(driver=driver)
        action_builder = ActionBuilder(driver=driver)

        if action:
            while True:
                x = action.x if hasattr(action, 'x') else 0
                y = action.y if hasattr(action, 'y') else 0
                scroll_x = action.scroll_x if hasattr(action, 'scroll_x') else 0
                scroll_y = action.scroll_y if hasattr(action, 'scroll_y') else 0
                text = action.text if hasattr(action, 'text') else ''
                path = handle_directory(directory_name='_screenshots', filename=f'cua_selenium', ext='png')

                match action.type:
                    case 'click':
                        self._logger.log(message="Action: click", color="blue", level="info")
                        driver.execute_script(f'window.scrollBy({x}, {y})')
                        action_chains.move_by_offset(xoffset=x, yoffset=y)
                        action_chains.perform()

                        if hasattr(action, 'button'):
                            match action.button:
                                case 'left':
                                    action_chains.click()
                                case 'right':
                                    action_chains.context_click()
                        action_chains.perform()
                        time.sleep(1)

                    case "scroll" | "move":
                        self._logger.log(message="Action: scroll", color="blue", level="info")
                        driver.execute_script(f'window.scrollBy({scroll_x}, {scroll_y})')
                        time.sleep(1)

                    case "keypress":
                        self._logger.log(message="Action: keypress", color="blue", level="info")
                        keys = action.keys
                        if keys:
                            for k in keys:
                                match k.lower():
                                    case "enter": action_chains.key_down(Keys.ENTER).perform()
                                    case "space": action_chains.key_down(Keys.SPACE).perform()
                                    case "select_all":
                                        if platform.system() == 'Darwin':
                                            action_chains.send_keys(Keys.COMMAND + "a").perform()
                                        else:
                                            action_chains.send_keys(Keys.CONTROL + "a").perform()
                                    case _:
                                        action_chains.key_down(Keys.SHIFT).send_keys(k).key_up(Keys.SHIFT).perform()
                        time.sleep(1)

                    case "type":
                        self._logger.log(message="Action: type", color="blue", level="info")
                        action_chains.send_keys(text).perform()
                        time.sleep(1)

                    case "wait":
                        self._logger.log(message="Action: wait", color="blue", level="info")
                        action_chains.pause(3)

                    case "screenshot":
                        self._logger.log(message="Action: screenshot", color="blue", level="info")
                        driver.save_screenshot(path)
                        time.sleep(1)

                    case _:
                        self._logger.log(message=f"Unrecognized action: {action}", level="warning", color="yellow")
                        return False

                with open(path, "rb") as image_file:
                    res, usage = None, None
                    if image_file:
                        screenshot_base64 = base64.b64encode(image_file.read()).decode("utf-8")
                        res, _, usage = self._run(screenshot=screenshot_base64)
                    else:
                        res, _, usage = self._run()

                    self._usage.aggregate(metrics=usage)
                    if not res:
                        usage.record_errors(type=ErrorType.API)
                        break

                    actions = [v for k, v in res.items() if k =="action"] if res else []
                    action = actions[0] if actions else None
                    if not action:
                        self._logger.log(message="No action found.", color="yellow", level="warning")
                        break
        else:
            self._usage.record_errors(type=ErrorType.TOOL)

        end_dt = datetime.datetime.now()
        self._usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        return res


    def run(self) -> Tuple[Dict[str, Any], None, UsageMetrics]:
        """Core function to execute the tool."""

        res = None
        try:
            res = self.invoke_playwright()
        except:
            self._call_ids = []
            self._calls = dict()
            self._response_ids = []
            res = self.invoke_selenium()

        return res, None, self._usage



REFLECT = "Here is the orignal prompt: {original_prompt}\nHere is the original response: {original_response}\nAnalyze the original prompt and repsonse, check for any pontential issue, and create an improved response."

INTEGRATE = "Here is the original prompt: {original_prompt}\nHere are responses: {responses}. Help integrate them as a single response."

parameter_sets = [
    {
        "temperature": 0.2,
        "top_p": 0.5,
        "max_tokens": 5000,
        "frequency_penalty": 0.5,
        "presence_penalty": 0.5,
        "stop": ["\n\n", "###"],
    },
    {
        "temperature": 0.7,
        "top_p": 0.8,
        "max_tokens": 8000,
        "frequency_penalty": 0.3,
        "presence_penalty": 0.3,
        "stop": ["\n\n"],
    },
    {
        "temperature": 1.0,
        "top_p": 0.95,
        "max_tokens": 12000,
        "frequency_penalty": 0.0,
        "presence_penalty": 0.0,
        "stop": [],
    }
]



from unittest.mock import patch


def test_draft():
    import versionhq as vhq
    from pydantic import BaseModel

    class Trip(BaseModel):
        name: str
        location: str
        description: str
        date: str
        cousine: str
        why_its_suitable: str


    with patch('builtins.input', return_value='Y'):
        task_graph = vhq.workflow(
            Trip,
            context="Planning a suprise day trip for my friend to celebrate her birthday. We live in CA and we like to have Korean food.",
            human=True
        )

        assert task_graph is not None
        assert [k == node.identifier and node.task and isinstance(node, vhq.Node) for k, node in task_graph.nodes.items()]
        assert [isinstance(edge.dependency_type, vhq.DependencyType) and isinstance(edge, vhq.Edge) for edge in task_graph.edges.values()]
        assert [k in task_graph.nodes.keys() and v.status == vhq.TaskStatus.NOT_STARTED for k, v in task_graph.nodes.items()]
        assert task_graph.should_reform == True
        assert task_graph.reform_trigger_event == vhq.ReformTriggerEvent.USER_INPUT


def test_condition_and_human_input():
    import versionhq as vhq

    task_a = vhq.Task(description="draft a message")
    task_b = vhq.Task(description="ask human for feedback")
    task_c = vhq.Task(description="send a message")

    tg = vhq.TaskGraph(directed=True)

    node_a = tg.add_task(task=task_a)
    node_b = tg.add_task(task=task_b)
    node_c = tg.add_task(task=task_c)

    def cond_method(): return True

    def cond_method_with_args(**kwargs):
        item = kwargs.get("item", None)
        return True if item else False

    complex_condition = vhq.Condition(
        methods={"0": cond_method, "1": cond_method_with_args, "2": cond_method_with_args},
        args={"2": dict(item="Hi!"), },
        type=vhq.ConditionType.AND
    )

    tg.add_dependency(
        node_a.identifier, node_b.identifier,
        dependency_type=vhq.DependencyType.FINISH_TO_START,
        required=False,
        condition=vhq.Condition(methods={ "0": cond_method, }),
    )

    edge = [v for v in tg.edges.values()][0]
    assert isinstance(edge, vhq.Edge)
    assert edge.required == False
    assert edge.dependency_type == vhq.DependencyType.FINISH_TO_START
    assert edge.condition == vhq.Condition(methods={ "0": cond_method, })
    assert edge.dependency_met() == True


    tg.add_dependency(
        node_b.identifier, node_c.identifier,
        dependency_type=vhq.DependencyType.FINISH_TO_START,
        required=False,
        condition=vhq.Condition(
            methods={ "0": cond_method, "1": cond_method_with_args, "2": complex_condition },
            type=vhq.ConditionType.AND
        ),
    )

    edge = [v for v in tg.edges.values()][1]
    assert isinstance(edge, vhq.Edge)
    assert edge.required == False
    assert edge.dependency_type == vhq.DependencyType.FINISH_TO_START
    assert edge.condition == vhq.Condition(methods={ "0": cond_method, "1": cond_method_with_args, "2": complex_condition }, type=vhq.ConditionType.AND)
    assert edge.dependency_met() == False

    setattr(tg, "should_reform", True)
    setattr(tg, "reform_trigger_event", vhq.ReformTriggerEvent.USER_INPUT)

    with patch.object(vhq.TaskGraph, "handle_reform", return_value=(vhq.TaskOutput(raw="mock"),{"a": vhq.TaskOutput(raw="mock")})) as mock_handle_reform:
        tg.activate()
        mock_handle_reform.assert_called_once()



def test_doc_an_formation_a():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task="Find the best trip destination this summer.",
        expected_outcome="a list of destinations and why it's suitable",
        context="planning a suprise trip for my friend", # optional
    )

    assert isinstance(network, vhq.AgentNetwork)
    assert network.members is not None # auto-generated agents as network members
    assert network.tasks is not None # auto-defined sub-tasks to achieve the main task goal


def test_doc_an_formation_b():
    import versionhq as vhq
    from pydantic import BaseModel

    class Outcome(BaseModel):
        destinations: list[str]
        why_suitable: list[str]


    network = vhq.form_agent_network(
        task="Find the best trip destination this summer.",
        expected_outcome=Outcome,
        context="planning a suprise trip for my friend", # optional
    )

    assert isinstance(network, vhq.AgentNetwork)
    assert network.members
    assert network.tasks


def test_doc_an_formation_c():
    import versionhq as vhq
    from pydantic import BaseModel

    my_agent = vhq.Agent(
        role="Travel Agent",
        goal="select best trip destination",
        knowledge_sources=[".....","url1",]
    )

    class Outcome(BaseModel):
        destinations: list[str]
        why_suitable: list[str]

    network = vhq.form_agent_network(
        task="Find the best trip destination this summer.",
        expected_outcome=Outcome,
        context="planning a suprise trip for my friend",
        agents=[my_agent,]
    )

    assert isinstance(network, vhq.AgentNetwork)
    assert [member for member in network.members if member.agent.role == my_agent.role]
    assert network.tasks


def test_doc_an_formation_d():
    import versionhq as vhq
    from pydantic import BaseModel

    my_agent = vhq.Agent(
        role="Travel Agent",
        goal="select best trip destination",
        knowledge_sources=[".....","url1",]
    )

    class Outcome(BaseModel):
        destinations: list[str]
        why_suitable: list[str]

    network = vhq.form_agent_network(
        task="Find the best trip destination this summer.",
        expected_outcome=Outcome,
        context="planning a suprise trip for my friend",
        agents=[my_agent,],
        formation=vhq.Formation.SUPERVISING
    )

    assert isinstance(network, vhq.AgentNetwork)
    assert [member for member in network.members if member.agent == my_agent]
    assert network.tasks
    assert network.formation == vhq.Formation.SUPERVISING


def test_doc_an_config_a():
    import versionhq as vhq

    network = vhq.AgentNetwork(
        members=[
            vhq.Member(
                agent=vhq.Agent(role="new member", goal="work in the network"),
                is_manager=False  # explicitly mentioned. Setting `True` makes this member a manager of the network.
            ),
        ]
    )
    assert isinstance(network.members[0].agent, vhq.Agent)


def test_doc_an_config_b():
    import versionhq as vhq

    network = vhq.AgentNetwork(
    members=[
        vhq.Member(agent=vhq.Agent(role="member 1", goal="work in the network")),
        vhq.Member(agent=vhq.Agent(role="member 2", goal="work in the network")),
        vhq.Member(agent=vhq.Agent(role="member 3", goal="work in the network")),
    ],
    formation=vhq.Formation.SQUAD,
    )

    assert network.formation == vhq.Formation.SQUAD


def test_doc_an_config_c():
    import versionhq as vhq

    network = vhq.AgentNetwork(
    members=[
        vhq.Member(agent=vhq.Agent(role="member 1", goal="work in the network"), tasks=[vhq.Task(description="Run a demo 1")]),
        vhq.Member(agent=vhq.Agent(role="member 2", goal="work in the network"), tasks=[vhq.Task(description="Run a demo")]),
        vhq.Member(agent=vhq.Agent(role="member 3", goal="work in the network")),
    ],
    )

    res, tg = network.launch()

    assert isinstance(res, vhq.TaskOutput)
    assert isinstance(tg, vhq.TaskGraph)


def test_doc_an_config_d():
    import versionhq as vhq

    network = vhq.AgentNetwork(
    members=[
        vhq.Member(agent=vhq.Agent(role="member 1", goal="work in the network"), tasks=[vhq.Task(description="Run a demo 1")]),
        vhq.Member(agent=vhq.Agent(role="member 2", goal="work in the network"), tasks=[vhq.Task(description="Run a demo 2")]),
        vhq.Member(agent=vhq.Agent(role="member 3", goal="work in the network")),
    ],
    process=vhq.TaskHandlingProcess.CONSENSUAL,
    consent_trigger=lambda x: True, # consent trigger event is a MUST for TaskHandlingProcess.CONSENSUAL
    )

    res, tg = network.launch()

    assert isinstance(res, vhq.TaskOutput)
    assert isinstance(tg, vhq.TaskGraph)



from typing import Any, Dict, List, Optional
from pydantic import BaseModel, ConfigDict, Field

from versionhq.knowledge.storage import KnowledgeStorage
from versionhq.knowledge.source import BaseKnowledgeSource


class Knowledge(BaseModel):
    """
    Knowlede class for collection of sources and setup for the vector store to query relevant context.
    """

    collection_name: Optional[str] = None
    sources: List[BaseKnowledgeSource] = Field(default_factory=list)
    storage: KnowledgeStorage = Field(default_factory=KnowledgeStorage)
    embedder_config: Optional[Dict[str, Any]] = None
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        collection_name: str,
        sources: List[BaseKnowledgeSource],
        embedder_config: Optional[Dict[str, Any]] = None,
        storage: Optional[KnowledgeStorage] = None,
        **data,
    ):
        super().__init__(**data)

        self.collection_name = collection_name
        self.sources = sources
        self.embedder_config = embedder_config
        self.storage = storage if storage else KnowledgeStorage(embedder_config=embedder_config, collection_name=self.collection_name)
        self.storage.initialize_knowledge_storage()

        for source in sources:
            source.storage = self.storage
            source.add()


    def query(self, query: List[str], limit: int = 3) -> List[Dict[str, Any]]:
        """
        Query across all knowledge sources to find the most relevant information.
        Returns the top_k most relevant chunks.
        """

        results = self.storage.search(query, limit)
        return results


    def _add_sources(self):
        for source in self.sources:
            source.storage = self.storage
            source.add()



import os
import datetime
from pathlib import Path


def handle_directory(directory_name: str = None, filename: str = None, ext: str = 'png') -> Path:
    """Creates and returns the absolute file path"""

    os.makedirs(directory_name, exist_ok=True)

    date = str(datetime.datetime.now().strftime('%j'))
    cwd = Path.cwd()
    DIRECTORY = cwd / f'{directory_name}/{filename}_{date}.{ext}'

    return DIRECTORY



import datetime
from typing import List, Dict, Any, Optional, Tuple

from versionhq.tool.gpt import openai_client
from versionhq.tool.gpt._enum import GPTFilterTypeEnum
from versionhq._utils import is_valid_enum, UsageMetrics, ErrorType


def is_valid_vector_store_id(id: str | list[str]) -> bool:
    if isinstance(id, list):
        for item in id:
            if not id.startswith("vs_"):
                return False
    else:
        return id.startswith("vs_")


class FilterSchema:
    class Filter:
        type: GPTFilterTypeEnum = GPTFilterTypeEnum.eq
        property: str = None
        value: str = None

        def __init__(self, **kwargs):
            for k, v in kwargs:
                if hasattr(self, k):
                    setattr(self, k, v)

        def _convert_to_schema(self) -> Dict[str, str] | None:
            return { "type": self.type, "property": self.property, "value": self.value } if self.property and self.value else None

    logic_type: str = "and"  # or
    filters: List[Filter] = list()
    filter_params: Dict[str, Any] = None

    def __init__(self, logic_type: str = None, filters: List[Filter] | Filter = None, filter_params: Dict[str, Any] = None, **kwargs):
        if logic_type == "and" | "or":
            self.logic_type = logic_type

        if filter_params:
            filter = FilterSchema.Filter()
            for k, v in filter_params.items():
                if k in FilterSchema.Filter.__dict__.keys():
                    if k == "type" and is_valid_enum(val=v, enum=GPTFilterTypeEnum):
                        setattr(filter, k, v if isinstance(v, str) else v.value)
                    else:
                        setattr(filter, k, v)
            self.filters.append(filter)

        if filters:
            match filters:
                case list():
                    if self.filters:
                        self.filters.extend(filters)
                    else:
                        self.filters = filters

                case FilterSchema.Filter():
                    if self.filters:
                        self.filters.append(filters)
                    else:
                        self.filters = [filters]

                case _:
                    pass

        if kwargs:
            for k, v in kwargs:
                if hasattr(self, k): setattr(self, k, v)

    @property
    def schema(self) -> Dict[str, Any] | None:
        if self.type and len(self.items) > 1:
            return {
                "type": self.type,
                "filters": [item._convert_to_schema() for item in self.items if isinstance(item, FilterSchema.Filter)]
            }
        elif self.filters:
            return self.filters[0]._convert_to_schema()
        else:
            return None


class GPTToolFileSearch:
    model: str = "gpt-4o"
    input: str = None
    vector_store_ids: List[str] = list()
    max_num_results: int = 2
    include: List[str] = ["output[*].file_search_call.search_results"]
    filters: Optional[FilterSchema] = None
    _usage: UsageMetrics = UsageMetrics()

    def __init__(
            self,
            input: str,
            vector_store_ids: str | List[str],
            model: str = None,
            max_num_results: int = None,
            include: List[str] = None,
            filters: FilterSchema | Dict[str, Any] = None,
            **kwargs,
        ):
        if not input or not vector_store_ids:
            return None

        if not is_valid_vector_store_id(id=vector_store_ids):
            return None

        self.input = input
        self.vector_store_ids = vector_store_ids if isinstance(vector_store_ids, list) else [vector_store_ids]
        self.model = model if model else self.model
        self.max_num_results = max_num_results if max_num_results else self.max_num_results
        self.include = include if include else self.include
        self.filters = filters if filters else None
        if kwargs:
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, v)


    def run(self) -> Tuple[str, List[Dict[str, Any]], UsageMetrics] | None:
        raw_res = ""
        annotations = list()
        usage = self._usage if self._usage else UsageMetrics()
        start_dt = datetime.datetime.now()

        try:
            res = openai_client.responses.create(**self.schema)
            if not res:
                usage.record_errors(ErrorType.TOOL)
            else:
                raw_res = res.output[1].content[0].text
                annotations = [{ "index": item.index, "file_id": item.file_id, "filename": item.filename }
                                for item in res.output[1].content[0].annotations]
                usage.record_token_usage(**res.usage.__dict__)

        except:
            usage.record_errors(ErrorType.TOOL)

        end_dt = datetime.datetime.now()
        usage.record_latency(start_dt=start_dt, end_dt=end_dt)
        self._usage = usage
        return raw_res, annotations, usage


    @property
    def tool_schema(self) -> Dict[str, Any]:
        if self.filters:
            return [
                {
                    "type": "file_search",
                    "vector_store_ids": self.vector_store_ids,
                    "max_num_results": self.max_num_results,
                    "filters": self.filters.schema,
                }
            ]
        else:
            return [
                {
                    "type": "file_search",
                    "vector_store_ids": self.vector_store_ids,
                    "max_num_results": self.max_num_results,
                }
            ]


    @property
    def schema(self) -> Dict[str, Any]:
        schema = dict(model=self.model, tools=self.tool_schema, input=self.input)
        return schema



import os

from versionhq.tool.composio.model import ComposioBaseTool
from versionhq.tool.composio.params import ComposioAppName, ComposioAuthScheme, ComposioStatus, ComposioAction

DEFAULT_MODEL_NAME = os.environ.get("DEFAULT_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")
TEST_CONNECTED_ACCOUNT_ID = os.environ.get("TEST_CONNECTED_ACCOUNT_ID", "f79b1a07-db14-4aa5-959e-5713ccd18de2")


def _test_connect_with_composio():
    composio = ComposioBaseTool(
        app_name=ComposioAppName.HUBSPOT,
        auth_scheme=ComposioAuthScheme.OAUTH2,
        connected_account_id=TEST_CONNECTED_ACCOUNT_ID
    )
    composio, status = composio._connect()
    assert composio.connected_account_id is not None
    assert status is not ComposioStatus.FAILED


def _test_execute_action_on_langchain():
    composio = ComposioBaseTool(
        app_name=ComposioAppName.HUBSPOT,
        auth_scheme=ComposioAuthScheme.OAUTH2,
        connected_account_id=TEST_CONNECTED_ACCOUNT_ID
    )
    composio, status = composio._connect()
    composio, res = composio.execute_composio_action_with_langchain(
        action_name=ComposioAction.HUBSPOT_CREATE_PIPELINE_STAGE,
        task_in_natural_language="create a demo pipeline"
    )
    assert status == ComposioStatus.ACTIVE.value
    assert composio.connected_account_id == TEST_CONNECTED_ACCOUNT_ID
    assert composio.tools is not None
    assert res is not None
    assert isinstance(res, str)



import uuid
import datetime
from enum import IntEnum
from typing import Dict, List
from typing_extensions import Self

from pydantic import BaseModel, UUID4, InstanceOf


class ErrorType(IntEnum):
    FORMAT = 1
    TOOL = 2
    API = 3
    OVERFITTING = 4
    HUMAN_INTERACTION = 5


class UsageMetrics(BaseModel):
    """A Pydantic model to manage token usage, errors, job latency."""

    id: UUID4 = uuid.uuid4() # stores task id or task graph id
    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    input_tokens: int = 0
    output_tokens: int = 0
    successful_requests: int = 0
    total_errors: int = 0
    error_breakdown: Dict[ErrorType, int] = dict()
    latency: float = 0.0  # in ms


    def record_token_usage(self, *args, **kwargs) -> None:
        """Records usage metrics from the raw response of the model."""

        if args:
            for item in args:
                match item:
                    case dict():
                        if hasattr(self, k):
                            setattr(self, k, int(getattr(self, k)) + int(v))
                    case UsageMetrics():
                        self = self.aggregate(metrics=item)
                    case _:
                        try:
                            self.completion_tokens += item.completion_tokens if hasattr(item, "completion_tokens") else 0
                            self.prompt_tokens += item.prompt_tokens if hasattr(item, "prompt_tokens") else 0
                            self.total_tokens += item.total_tokens if hasattr(item, "total_tokens") else 0
                            self.input_tokens += item.input_tokens if hasattr(item, "input_tokens") else 0
                            self.output_tokens += item.output_tokens if hasattr(item, "output_tokens") else 0
                        except:
                            pass
        if kwargs:
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, int(getattr(self, k)) + int(v))


    def record_errors(self, type: ErrorType = None) -> None:
        self.total_errors += 1
        if type:
            if type in self.error_breakdown:
                self.error_breakdown[type] += 1
            else:
                self.error_breakdown[type] = 1


    def record_latency(self, start_dt: datetime.datetime, end_dt: datetime.datetime) -> None:
        self.latency += round((end_dt - start_dt).total_seconds() * 1000, 3)


    def aggregate(self, metrics: InstanceOf["UsageMetrics"]) -> Self:
        if not metrics:
            return self

        self.total_tokens += metrics.total_tokens
        self.prompt_tokens += metrics.prompt_tokens
        self.completion_tokens += metrics.completion_tokens
        self.input_tokens += metrics.input_tokens
        self.output_tokens += metrics.output_tokens
        self.successful_requests += metrics.successful_requests
        self.total_errors += metrics.total_errors
        self.latency += metrics.latency
        self.latency = round(self.latency, 3)

        if metrics.error_breakdown:
            for k, v in metrics.error_breakdown.items():
                if self.error_breakdown and k in self.error_breakdown:
                    self.error_breakdown[k] += int(v)
                else:
                    self.error_breakdown.update({ k: v })

        return self



import os
from typing import Any, Dict, List

from mem0 import MemoryClient

from versionhq.storage.base import Storage


class Mem0Storage(Storage):
    """
    Extends Storage to handle embedding and searching across entities using Mem0.
    """

    def __init__(self, type, agent=None, user_id=None):
        """
        Create a memory client using API keys and other config.
        """

        super().__init__()

        if type not in ["user", "stm", "ltm", "entities"]:
            raise ValueError("Invalid type for Mem0Storage. Must be 'user' or 'agent'.")

        self.memory_type = type
        self.agent= agent
        self.memory_config = agent.memory_config

        user_id = user_id if user_id else self._get_user_id()
        if type == "user" and not user_id:
            raise ValueError("User ID is required for user memory type")

        config = self.memory_config.get("config", {})
        mem0_api_key = os.environ.get("MEM0_API_KEY", config.get("api_key"))
        mem0_org_id = config.get("org_id")
        mem0_project_id = config.get("project_id")

        if mem0_org_id and mem0_project_id:
            self.memory = MemoryClient(api_key=mem0_api_key, org_id=mem0_org_id, project_id=mem0_project_id)
        else:
            self.memory = MemoryClient(api_key=mem0_api_key)


    # def _sanitize_role(self, role: str) -> str:
    #     """
    #     Sanitizes agent roles to ensure valid directory names.
    #     """
    #     return role.replace("\n", "").replace(" ", "_").replace("/", "_")


    def save(self, value: Dict[str, Any] | str, metadata: Dict[str, Any]) -> None:
        user_id = self._get_user_id()
        agent_name = self._get_agent_name()

        if self.memory_type == "user":
            self.memory.add(value, user_id=user_id, metadata={**metadata})

        elif self.memory_type == "stm":
            agent_name = self._get_agent_name()
            self.memory.add(value, agent_id=agent_name, metadata={"type": "stm", **metadata})

        elif self.memory_type == "ltm":
            agent_name = self._get_agent_name()
            self.memory.add(value, agent_id=agent_name, infer=False, metadata={"type": "ltm", **metadata})

        elif self.memory_type == "entities":
            entity_name = self._get_agent_name()
            self.memory.add(value, user_id=entity_name, metadata={"type": "entity", **metadata})


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35) -> List[Any]:
        params = {"query": query, "limit": limit}

        if self.memory_type == "user":
            user_id = self._get_user_id()
            params["user_id"] = user_id

        elif self.memory_type == "stm":
            agent_name = self._get_agent_name()
            params["agent_id"] = agent_name
            params["metadata"] = {"type": "stm"}

        elif self.memory_type == "ltm":
            agent_name = self._get_agent_name()
            params["agent_id"] = agent_name
            params["metadata"] = {"type": "ltm"}

        elif self.memory_type == "entities":
            agent_name = self._get_agent_name()
            params["agent_id"] = agent_name
            params["metadata"] = {"type": "entity"}

        results = self.memory.search(**params)
        return [r for r in results if r["score"] >= score_threshold]


    def _get_user_id(self):
        if self.memory_type == "user":
            if hasattr(self, "memory_config") and self.memory_config is not None:
                return self.memory_config.get("config", {}).get("user_id")
            else:
                return None
        return None


    def _get_agent_name(self):
        agents = self.agents if self.agents else []
        agents = "_".join([item.key for item in agents])
        return agents



from typing import List, Optional, Dict
from typing_extensions import Self

from pydantic import InstanceOf, Field

from versionhq.agent.model import Agent
from versionhq.task.model import Task
from versionhq.task_graph.model import TaskGraph, Node, DependencyType
from versionhq._prompt.model import Prompt
from versionhq._prompt.constants import REFLECT, INTEGRATE, parameter_sets


class PromptFeedbackGraph(TaskGraph):
    """A Pydantic class to handle auto prompt feedback cycle."""

    _times_iteration: int = 0
    user_prompts: Optional[Dict[str, str]] = Field(default_factory=dict) # { "0": "...", "1": "..."}
    dev_prompts: Optional[Dict[str, str]] = Field(default_factory=dict)
    prompts: Optional[Dict[str, InstanceOf[Prompt]]] = Field(default_factory=dict)


    def __init__(self, prompt: InstanceOf[Prompt] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if prompt:
            user_prompt, dev_prompt, _ = prompt.format_core()
            self.prompts = { self.key: prompt }
            self.user_prompts = { self.key: user_prompt }
            self.dev_prompts = { self.key: dev_prompt }


    def _fetch_latest_prompt(self) -> InstanceOf[Prompt] | None:
        return self.prompts[self.key] if self.key in self.prompts else None


    def _generate_agents(self) -> List[Agent] | None:
        agents = []
        prompt = self._fetch_latest_prompt()

        if not prompt:
            return None

        agent = prompt.agent
        agent_params = agent.model_dump(exclude={"id", "llm", "llm_config", "self_learning"})
        for params in parameter_sets:
            agent = Agent(**agent_params, llm=agent.llm.model, llm_config={**params}, self_learning=True)
            agents.append(agent)
        return agents


    def _reflect(self, original_response: str) -> Task:
        description = REFLECT.format(original_prompt=self.original_prompt, original_response=original_response)
        return Task(description=description)


    def set_up_graph(self, **attributes) -> Self:
        """Sets up a TaskGraph object with nodes and edges."""

        prompt = self._fetch_latest_prompt()
        base_task = prompt.task if prompt else None
        base_agent = prompt.agent if prompt else None

        if not base_task or not base_agent:
            return None

        agents = self._generate_agents()
        if not agents:
            return None

        self.concl_response_schema = base_task.response_schema
        base_agent.callbacks.append(self._reflect)
        init_node = Node(task=base_task, assigned_to=base_agent)
        self.add_node(init_node)

        final_task = Task(description=INTEGRATE.format(original_prompt=self.original_prompt, responses=""))
        final_node = Node(task=final_task, agent=base_agent)
        self.add_node(node=final_node)

        for agent in agents:
            node = Node(task=base_task, assigned_to=agent)
            self.add_node(node=node)
            self.add_dependency(source=init_node.identifier, target=node.identifier, dependency_type=DependencyType.FINISH_TO_START, required=True)
            self.add_dependency(source=node.identifier, target=final_node.identifier, dependency_type=DependencyType.FINISH_TO_START, required=True)

        if attributes:
            for k, v in attributes.items():
                if hasattr(self, k):
                    setattr(self, k, v)

        return self

    @property
    def index(self) -> str:
        """Returns an index to add new item."""
        return str(len([k for k in self.user_prompts.keys()]))

    @property
    def original_prompt(self) -> str:
        return str(self.user_prompts["0"]) + str(self.dev_prompts["0"])

    @property
    def key(self):
        return str(self._times_iteration)



from unittest.mock import patch


def test_eval():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        topics: list[str]
        reasons: list[str]

    task = vhq.Task(
        description="Research a topic to teach a kid aged 6 about math.",
        response_schema=CustomOutput,
        should_evaluate=True,
        eval_criteria=["Uniquness", "Fit to audience",],
        # fsls=[""]
    )

    with patch.object(vhq.Evaluation, "_draft_fsl_prompt", return_value="test") as draft_fsl_prompt:
        res = task.execute()
        assert isinstance(res.evaluation, vhq.Evaluation)
        assert [isinstance(item, vhq.EvaluationItem) and item.criteria in task.eval_criteria for item in res.evaluation.items]
        assert res.evaluation.aggregate_score is not None
        assert res.evaluation.suggestion_summary
        assert res.usage is not None

        draft_fsl_prompt.assert_called_once_with(task_description=task.description)


def test_eval_with_fsls():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        topics: list[str]
        reasons: list[str]

    task = vhq.Task(
        description="Research a topic to teach a kid aged 6 about math.",
        response_schema=CustomOutput,
        should_evaluate=True,
        eval_criteria=["Uniquness", "Fit to audience",],
        fsls=["Start by explaining that math is all around us and helps us solve problems in our daily lives. Today, we're going to learn about adding and subtracting using objects.", "Focus on complex concepts like seconds and milliseconds."]
    )

    res = task.execute()
    assert isinstance(res.evaluation, vhq.Evaluation)
    assert [isinstance(item, vhq.EvaluationItem) and item.criteria in task.eval_criteria for item in res.evaluation.items]
    assert res.evaluation.aggregate_score is not None
    assert res.evaluation.suggestion_summary is not None



BACKSTORY_FULL="""You are an expert {role} highly skilled in {skills}. You have abilities to query relevant information from the given knowledge sources and use tools such as {tools}. Leveraging these, you will identify competitive solutions to achieve the following goal: {goal}."""


BACKSTORY_SHORT="""You are an expert {role} with relevant skills and abilities to query relevant information from the given knowledge sources. Leveraging these, you will identify competitive solutions to achieve the following goal: {goal}."""



import base64

def convert_img_url(img_url: str) -> str | None:
    try:
        with open(img_url, "rb") as file:
            content = file.read()
            if content:
                encoded_file = base64.b64encode(content).decode("utf-8")
                img_url = f"data:image/jpeg;base64,{encoded_file}"
                return img_url

            else: return None

    except:
        return None



import os
import uuid
from typing import Any, Dict, List, Optional, TypeVar, Callable, Type, Tuple
from typing_extensions import Self
from dotenv import load_dotenv

from pydantic import UUID4, BaseModel, Field, InstanceOf, PrivateAttr, model_validator, field_validator
from pydantic_core import PydanticCustomError

from versionhq.agent.rpm_controller import RPMController
from versionhq.tool.model import Tool, ToolSet, BaseTool
from versionhq.knowledge.model import BaseKnowledgeSource, Knowledge
from versionhq.memory.model import ShortTermMemory, LongTermMemory, UserMemory
from versionhq._utils import Logger, process_config, is_valid_url, ErrorType, UsageMetrics


load_dotenv(override=True)
T = TypeVar("T", bound="Agent")


# @track_agent()
class Agent(BaseModel):
    """
    A Pydantic class to store an agent object. Agents must have `role` and `goal` to start.
    """

    __hash__ = object.__hash__
    _rpm_controller: Optional[RPMController] = PrivateAttr(default=None)
    _request_within_rpm_limit: Any = PrivateAttr(default=None)
    _times_executed: int = PrivateAttr(default=0)
    _logger_config: Dict[str, Any] = PrivateAttr(default=dict(verbose=True, info_file_save=True))

    api_key: Optional[str] = Field(default=None)
    self_learn: bool = Field(default=False)
    config: Optional[Dict[str, Any]] = Field(default=None, exclude=True, description="values to add to the Agent class")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    role: str = Field(description="required. agent's role")
    goal: Optional[str] = Field(default=None)
    backstory: Optional[str] = Field(default=None, description="developer prompt to the llm")
    skills: Optional[List[str]] = Field(default_factory=list, description="list up the agent's tangible skills in natural language")
    tools: Optional[List[Any]] = Field(default_factory=list)

    # knowledge
    knowledge_sources: Optional[List[BaseKnowledgeSource | Any]] = Field(default=None)
    embedder_config: Optional[Dict[str, Any]] = Field(default=None, description="embedder configuration for knowledge sources")
    _knowledge: Optional[Knowledge] = PrivateAttr(default=None)

    # memory
    with_memory: bool = Field(default=False, description="whether to use memories during the task execution")
    memory_config: Optional[Dict[str, Any]] = Field(default=None, description="memory config. needs to store user_id for UserMemory to work")
    short_term_memory: Optional[InstanceOf[ShortTermMemory]] = Field(default=None)
    long_term_memory: Optional[InstanceOf[LongTermMemory]] = Field(default=None)
    user_memory: Optional[InstanceOf[UserMemory]] = Field(default=None)

    # prompting
    use_developer_prompt: Optional[bool] = Field(default=True, description="Use developer prompt when calling the llm")
    developer_propmt_template: Optional[str] = Field(default=None, description="abs. file path to developer prompt template")
    user_prompt_template: Optional[str] = Field(default=None, description="abs. file path to user prompt template")

    # task execution rules
    networks: Optional[List[Any]] = Field(default_factory=list, description="store a list of agent networks that the agent belongs to as a member")
    allow_delegation: bool = Field(default=False, description="whether to delegate the task to another agent")
    max_retry_limit: int = Field(default=2, description="max. number of task retries when an error occurs")
    maxit: Optional[int] = Field(default=25, description="max. number of total optimization loops conducted when an error occurs")
    callbacks: Optional[List[Callable]] = Field(default_factory=list, description="callback functions to execute after any task execution")

    # llm settings cascaded to the LLM model
    llm: Any = Field(default=None, description="store LLM object")
    func_calling_llm: Any = Field(default=None, description="store LLM object")
    respect_context_window: bool = Field(default=True, description="keep messages under the context window size")
    max_execution_time: Optional[int] = Field(default=None, description="max. task execution time in seconds")
    max_rpm: Optional[int] = Field(default=None, description="max. number of requests per minute")
    llm_config: Optional[Dict[str, Any]] = Field(default=None, description="other llm config cascaded to the LLM class")

    # # cache, error, ops handling
    # formatting_errors: int = Field(default=0, description="number of formatting errors.")

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="before")
    @classmethod
    def process_model_config(cls, values: Dict[str, Any]) -> None:
        return process_config(values_to_update=values, model_class=cls)


    @model_validator(mode="after")
    def validate_required_fields(self) -> Self:
        required_fields = ["role",]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError(f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_llm(self) -> Self:
        """
        Set up `llm` and `func_calling_llm` as valid LLM objects using the given kwargs.
        """
        from versionhq.llm.model import LLM, DEFAULT_MODEL_NAME

        self.llm = self._convert_to_llm_object(llm=self.llm)

        func_calling_llm = self.func_calling_llm if self.func_calling_llm else self.llm if self.llm else None
        func_calling_llm = self._convert_to_llm_object(llm=func_calling_llm)
        if func_calling_llm._supports_function_calling():
            self.func_calling_llm = func_calling_llm
        elif self.llm._supports_function_calling():
            self.func_calling_llm = self.llm
        else:
            self.func_calling_llm = self._convert_to_llm_object(llm=LLM(model=DEFAULT_MODEL_NAME))
        return self


    @model_validator(mode="after")
    def set_up_tools(self) -> Self:
        """
        Similar to the LLM set up, when the agent has tools, we will declare them using the Tool class.
        """
        from versionhq.tool.rag_tool import RagTool
        from versionhq.tool.gpt.web_search import GPTToolWebSearch
        from versionhq.tool.gpt.file_search import GPTToolFileSearch
        from versionhq.tool.gpt.cua import GPTToolCUA

        if not self.tools:
            return self

        tool_list = []
        for item in self.tools:
            match item:
                case RagTool() | BaseTool() | GPTToolCUA() | GPTToolFileSearch() | GPTToolWebSearch():
                    tool_list.append(item)

                case Tool():
                    if item.func is not None:
                        tool_list.append(item)

                case ToolSet():
                    if item.tool and item.tool.func is not None:
                        tool_list.append(item)

                case dict():
                    if "func" in item:
                        tool = Tool(func=item["func"])
                        for k, v in item.items():
                            if k in Tool.model_fields.keys() and k != "func" and  v is not None:
                                setattr(tool, k, v)
                        tool_list.append(tool)

                case func if callable(func):
                    tool = Tool(func=item)
                    tool_list.append(tool)

                case _:
                    if hasattr(item, "__base__") and (item.__base__ == BaseTool or item.__base__ == RagTool or item.__base__ == Tool):
                        tool_list.append(item)
                    else:
                        Logger(**self._logger_config, filename=self.key).log(level="error", message=f"Tool {str(item)} is missing a function.", color="red")
                        raise PydanticCustomError("invalid_tool", f"The tool {str(item)} is missing a function.", {})

        self.tools = tool_list
        return self


    @model_validator(mode="after")
    def set_up_backstory(self) -> Self:
        """
        Set up the backstory using a templated BACKSTORY when the backstory is None
        """

        if self.backstory is None:
            from versionhq.agent.TEMPLATES.Backstory import BACKSTORY_FULL, BACKSTORY_SHORT
            backstory = ""
            skills = ", ".join([item for item in self.skills]) if self.skills else ""
            tools = ", ".join([item.name for item in self.tools if hasattr(item, "name") and item.name is not None]) if self.tools else ""
            role = self.role.lower()
            goal = self.goal.lower() if self.goal else ""

            if self.tools or self.skills:
                backstory = BACKSTORY_FULL.format(role=role, goal=goal, skills=skills, tools=tools)
            else:
                backstory = BACKSTORY_SHORT.format(role=role, goal=goal)

            self.backstory = backstory

        return self


    @model_validator(mode="after")
    def set_up_rpm(self) -> Self:
        """
        Set up RPM controller.
        """
        if self.max_rpm:
            self._rpm_controller = RPMController(max_rpm=self.max_rpm, _current_rpm=0)

        return self


    @model_validator(mode="after")
    def set_up_knowledge(self) -> Self:
        from versionhq.knowledge.source import BaseKnowledgeSource, StringKnowledgeSource, TextFileKnowledgeSource, CSVKnowledgeSource, ExcelKnowledgeSource, JSONKnowledgeSource
        from versionhq.knowledge.source_docling import DoclingSource

        if self.knowledge_sources:
            try:
                collection_name = self.key
                knowledge_sources = []
                docling_fp, txt_fp, json_fp, excel_fp, csv_fp, pdf_fp = [], [], [], [], [], []
                str_cont = ""

                for item in self.knowledge_sources:
                    if isinstance(item, BaseKnowledgeSource):
                        knowledge_sources.append(item)

                    elif isinstance(item, str) and "http" in item and is_valid_url(url=item) == True:
                        docling_fp.append(item)

                    elif isinstance(item, str):
                        match  os.path.splitext(item)[1]:
                            case ".txt": txt_fp.append(item)
                            case ".json": json_fp.append(item)
                            case ".xls" | ".xlsx": excel_fp.append(item)
                            case ".pdf": pdf_fp.append(item)
                            case ".csv": csv_fp.append(item)
                            case _: str_cont += str(item)

                    else:
                        str_cont += str(item)

                if docling_fp: knowledge_sources.append(DoclingSource(file_paths=docling_fp))
                if str_cont: knowledge_sources.append(StringKnowledgeSource(content=str_cont))
                if txt_fp: knowledge_sources.append(TextFileKnowledgeSource(file_paths=txt_fp))
                if csv_fp: knowledge_sources.append(CSVKnowledgeSource(file_path=csv_fp))
                if excel_fp: knowledge_sources.append(ExcelKnowledgeSource(file_path=excel_fp))
                if json_fp: knowledge_sources.append(JSONKnowledgeSource(file_paths=json_fp))

                self._knowledge = Knowledge(sources=knowledge_sources, embedder_config=self.embedder_config, collection_name=collection_name)

            except Exception as e:
                Logger(**self._logger_config, filename=self.key).log(level="warning", message=f"We cannot find the format for the source. Add BaseKnowledgeSource objects instead. {str(e)}", color="yellow")

        return self


    @model_validator(mode="after")
    def set_up_memory(self) -> Self:
        """
        Set up memories: stm, ltm, and um
        """

        # if self.with_memory == True:
        self.long_term_memory = self.long_term_memory if self.long_term_memory else LongTermMemory()
        self.short_term_memory = self.short_term_memory if self.short_term_memory else ShortTermMemory(agent=self, embedder_config=self.embedder_config)

        if hasattr(self, "memory_config") and self.memory_config is not None:
            user_id = self.memory_config.get("user_id", None)
            if user_id:
                self.user_memory = self.user_memory if self.user_memory else UserMemory(agent=self, user_id=user_id)
        else:
            self.user_memory = None

        return self


    def _convert_to_llm_object(self, llm: Any = None): # returns LLM object
        """
        Convert the given value to LLM object.
        When `llm` is dict or self.llm_config is not None, add these values to the LLM object after validating them.
        """
        from versionhq.llm.model import LLM, DEFAULT_MODEL_NAME

        llm = llm if llm else self.llm if self.llm else DEFAULT_MODEL_NAME

        if not llm:
            pass

        match llm:
            case LLM():
                return self._set_llm_params(llm_obj=llm, config=self.llm_config)

            case str():
                llm = LLM(model=llm)
                return self._set_llm_params(llm_obj=llm, config=self.llm_config)

            case dict():
                model_name = llm.pop("model_name", llm.pop("deployment_name", str(llm)))
                llm_obj = LLM(model=model_name if model_name else DEFAULT_MODEL_NAME)
                config = llm.update(self.llm_config) if self.llm_config else llm
                return self._set_llm_params(llm_obj, config=config)

            case _:
                model_name = (getattr(self.llm, "model_name") or getattr(self.llm, "deployment_name") or str(self.llm))
                llm = LLM(model=model_name if model_name else DEFAULT_MODEL_NAME)
                llm_params = {
                    "timeout": getattr(llm, "timeout", self.max_execution_time),
                    "callbacks": getattr(llm, "callbacks", None),
                    "llm_config": getattr(llm, "llm_config", None),
                    "base_url": getattr(llm, "base_url", None),
                }
                config = llm_params.update(self.llm_config) if self.llm_config else llm_params
                return self._set_llm_params(llm_obj=llm, config=config)


    def _set_llm_params(self, llm_obj, config: Dict[str, Any] = None):  # returns LLM object
        """
        Add valid params to the LLM object.
        """
        from versionhq.llm.model import LLM, DEFAULT_CONTEXT_WINDOW_SIZE
        from versionhq.llm.llm_vars import PROVIDERS

        llm = llm_obj if isinstance(llm_obj, LLM) else None

        if not llm: return

        if llm.provider is None:
            provider_name = llm.model.split("/")[0]
            valid_provider = provider_name if provider_name in PROVIDERS else None
            llm.provider = valid_provider

        if self.callbacks:
            llm.callbacks = self.callbacks
            llm._set_callbacks(llm.callbacks)

        if self.respect_context_window == False:
            llm.context_window_size = DEFAULT_CONTEXT_WINDOW_SIZE

        llm.timeout = self.max_execution_time if llm.timeout is None else llm.timeout

        if config:
            llm.llm_config = {k: v for k, v in config.items() if v or v == False}
            llm.setup_config()

        return llm


    def _train(self) -> Self:
        """
        Fine-tuned the base model using OpenAI train framework.
        """
        from versionhq.llm.model import LLM

        if not isinstance(self.llm, LLM):
            pass


    def _invoke(
        self,
        messages: List[Dict[str, str]] = None,
        response_format: Optional[Dict[str, Any]] = None,
        tools: Optional[List[InstanceOf[Tool]| InstanceOf[ToolSet] | Type[Tool]]] = None,
        tool_res_as_final: bool = False,
        file: str = None, # absolute path to the content file (for multimodal use)
        ) -> Tuple[str, UsageMetrics]:
        """Calls LLM."""

        task_execution_counter = 0
        iterations = 0
        raw_response = None
        usage = UsageMetrics()

        try:
            if self._rpm_controller and self.max_rpm:
                self._rpm_controller.check_or_wait()

            Logger(**self._logger_config, filename=self.key).log(level="info", message=f"Messages sent to the model: {messages}", color="blue")

            if tool_res_as_final:
                raw_response = self.func_calling_llm.call(messages=messages, tools=tools, tool_res_as_final=True, file=file)
                usage.record_token_usage(*self.func_calling_llm._usages)
            else:
                raw_response = self.llm.call(messages=messages, response_format=response_format, tools=tools, file=file)
                usage.record_token_usage(*self.llm._usages)

            task_execution_counter += 1
            Logger(**self._logger_config, filename=self.key).log(level="info", message=f"Agent response: {raw_response}", color="green")
            return raw_response, usage

        except Exception as e:
            usage.record_errors(type=ErrorType.API)
            Logger(**self._logger_config, filename=self.key).log(level="error", message=f"An error occured. The agent will retry: {str(e)}", color="red")

            while not raw_response and task_execution_counter <= self.max_retry_limit:
                while (not raw_response or raw_response == "" or raw_response is None) and iterations < self.maxit:
                    if self.max_rpm and self._rpm_controller:
                        self._rpm_controller.check_or_wait()

                    raw_response = self.llm.call(messages=messages, response_format=response_format, tools=tools)
                    usage.record_token_usage(*self.llm._usages)
                    iterations += 1

                task_execution_counter += 1
                Logger(**self._logger_config, filename=self.key).log(level="info", message=f"Agent #{task_execution_counter} response: {raw_response}", color="green")
                return raw_response, usage

            if not raw_response:
                Logger(**self._logger_config, filename=self.key).log(level="error", message="Received None or empty response from the model", color="red")
                raise ValueError("Invalid response from LLM call - None or empty.")


    def _update_llm(self, llm: Any = None, llm_config: Optional[Dict[str, Any]] = None) -> Self:
        """
        Updates llm and llm_config of the exsiting agent. (Other conditions will remain the same.)
        """

        if not llm and not llm_config:
            Logger(**self._logger_config, filename=self.key).log(level="error", message="Missing llm or llm_config values to update", color="red")
            pass

        self.llm = llm
        if llm_config:
            if self.llm_config:
                self.llm_config.update(llm_config)
            else:
                self.llm_config = llm_config

        return self.set_up_llm()


    def _sort_tools(self, task = None) -> Tuple[List[Any], List[Any], List[Any]]:
        """Sorts agent and task tools by class."""

        from versionhq.tool.rag_tool import RagTool
        from versionhq.tool.gpt.web_search import GPTToolWebSearch
        from versionhq.tool.gpt.file_search import GPTToolFileSearch
        from versionhq.tool.gpt.cua import GPTToolCUA

        all_tools = []
        if task: all_tools = task.tools + self.tools if task.can_use_agent_tools else task.tools
        else: all_tools = self.tools

        rag_tools, gpt_tools, tools = [], [], []
        if all_tools:
            for item in all_tools:
                match item:
                    case RagTool():
                        rag_tools.append(item)

                    case GPTToolCUA() | GPTToolFileSearch() | GPTToolWebSearch():
                        gpt_tools.append(item)

                    case Tool() | BaseTool() | ToolSet():
                        tools.append(item)

        return rag_tools, gpt_tools, tools


    def _handle_gpt_tools(self, gpt_tools: list[Any] = None) -> Any: # TaskOutput or None
        """Generates k, v pairs from multiple GPT tool results and stores them in TaskOutput class."""

        from versionhq.task.model import TaskOutput
        from versionhq._utils import UsageMetrics

        if not gpt_tools:
            return None

        tool_res = dict()
        annotation_set = dict()
        total_usage = UsageMetrics()

        for i, item in enumerate(gpt_tools):
            raw, annotations, usage = item.run()
            tool_res.update({ str(i): raw })

            if annotations:
                annotation_set.update({ str(i): annotations })
            total_usage.aggregate(metrics=usage)

        res = TaskOutput(raw=str(tool_res), tool_output=tool_res, usage=total_usage, annotations=annotation_set)
        return res


    def update(self, **kwargs) -> Self:
        """
        Update the existing agent. Address variables that require runnning set_up_x methods first, then update remaining variables.
        """

        if not kwargs:
            Logger(**self._logger_config, filename=self.key).log(level="error", message="Missing values to update", color="red")
            return self

        for k, v in kwargs.items():
            match k:
                case "tools":
                    self.tools = kwargs.get(k, self.tools)
                    self.set_up_tools()

                case "role" | "goal":
                    self.role = kwargs.get("role", self.role)
                    self.goal = kwargs.get("goal", self.goal)
                    if not self.backstory:
                        self.set_up_backstory()

                    if self.backstory:
                        self.backstory += f"new role: {self.role}, new goal: {self.goal}"

                case "max_rpm":
                    self.max_rpm = kwargs.get(k, self.max_rpm)
                    self.set_up_rpm()

                case "knowledge_sources":
                    self.knowledge_sources = kwargs.get("knowledge_sources", self.knowledge_sources)
                    self.set_up_knowledge()

                case "with_memory" | "memory_config":
                    self.with_memory = kwargs.get("with_memory", self.with_memory)
                    self.memory_config = kwargs.get("memory_config", self.memory_config)
                    self.set_up_memory()

                case "llm" | "llm_config":
                    self.llm = kwargs.get("llm", self.llm)
                    self.llm_config = kwargs.get("llm_config", self.llm_config)
                    self._update_llm(llm=self.llm, llm_config=self.llm_config)

                case _:
                    try:
                        setattr(self, k, v)
                    except Exception as e:
                        Logger(**self._logger_config, filename=self.key).log(level="error", message=f"Failed to update the field: {k} We'll skip it. Error: {str(e)}", color="red")
                        pass

        return self


    def start(
            self,
            context: Any = None,
            tool_res_as_final: bool = False,
            image: str = None,
            file: str = None,
            audio: str = None
        ) -> Any:
        """
        Defines and executes a task, then returns TaskOutput object with the generated task.
        """

        from versionhq.task.model import Task

        if not self.role:
            return None, None

        _, gpt_tools, _ = self._sort_tools()

        if gpt_tools and tool_res_as_final == True:
            res = self._handle_gpt_tools(gpt_tools=gpt_tools)
            return res

        class Output(BaseModel):
            result: str
            steps: list[str]

        task = Task(
            description=f"Generate a simple result in a sentence to achieve the goal: {self.goal if self.goal else self.role}. If needed, list up necessary steps in concise manner.",
            response_schema=Output,
            tool_res_as_final=tool_res_as_final,
            image=image, #REFINEME - query memory/knowledge or self create
            file=file,
            audio=audio,
            can_use_agent_tools=True if self.tools else False,
        )
        res = task.execute(agent=self, context=context)
        return res


    def execute_task(self, task, context: Optional[Any] = None) -> Tuple[str, str, Any, UsageMetrics]:
        """Handling task execution."""

        from versionhq._prompt.model import Prompt
        from versionhq.task.model import Task

        task: InstanceOf[Task] = task
        rag_tools, gpt_tools, tools = self._sort_tools(task=task)
        raw_response = ""
        user_prompt, dev_prompt = "", ""
        usage = UsageMetrics(id=task.id)
        file = task.audio if task.is_multimodal and task.audio else task.image if task.is_multimodal and task.image else task.file if task.is_multimodal and task.file else None

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller._reset_request_count()

        if task.tool_res_as_final == True and gpt_tools:
            self._times_executed += 1
            res = self._handle_gpt_tools(gpt_tools=gpt_tools)
            return user_prompt, dev_prompt, res, res.usage

        user_prompt, dev_prompt, messages = Prompt(task=task, agent=self, context=context).format_core(rag_tools=rag_tools, gpt_tools=gpt_tools)

        try:
            self._times_executed += 1
            raw_response, usage = self._invoke(
                messages=messages,
                response_format=task._structure_response_format(model_provider=self.llm.provider),
                tools=tools,
                tool_res_as_final=task.tool_res_as_final,
                file=file,
            )

        except Exception as e:
            self._times_executed += 1
            Logger(**self._logger_config, filename=self.key).log(level="error", message=f"The agent failed to execute the task. Error: {str(e)}", color="red")
            user_prompt, dev_prompt, raw_response, usage = self.execute_task(task, context)

            if self._times_executed > self.max_retry_limit:
                Logger(**self._logger_config, filename=self.key).log(level="error", message=f"Max retry limit has exceeded.", color="red")
                raise e

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller.stop_rpm_counter()

        return user_prompt, dev_prompt, raw_response, usage


    @property
    def key(self):
        """
        A key to identify an agent. Used in storage, logging, and other recodings.
        """
        sanitized_role = self.role.lower().replace(" ", "-").replace("/", "").replace("{", "").replace("}", "").replace("\n", "")[0: 16]
        return f"{str(self.id)}-{sanitized_role}"


    def __repr__(self):
        return f"Agent(role={self.role}, id={str(self.id)}"

    def __str__(self):
        return super().__str__()



import uuid
from versionhq._utils import UsageMetrics


def test_activate():
    import versionhq as vhq

    task_graph = vhq.TaskGraph(directed=False, should_reform=True)

    task_a = vhq.Task(description="Research Topic")
    task_b = vhq.Task(description="Outline Post")
    task_c = vhq.Task(description="Write First Draft")

    node_a = task_graph.add_task(task=task_a)
    node_b = task_graph.add_task(task=task_b)
    node_c = task_graph.add_task(task=task_c)

    task_graph.add_dependency(
        source=node_a.identifier, target=node_b.identifier,
        dependency_type=vhq.DependencyType.FINISH_TO_START, weight=5, description="B depends on A"
    )
    task_graph.add_dependency(
        source=node_a.identifier, target=node_c.identifier,
        dependency_type=vhq.DependencyType.FINISH_TO_FINISH, lag=1, required=False, weight=3
    )
    critical_path, duration, paths = task_graph.find_critical_path()
    assert isinstance(task_graph, vhq.TaskGraph)
    assert [type(k) == uuid.uuid4 and isinstance(v, vhq.Node) and isinstance(v.task, vhq.Task) for k, v in task_graph.nodes.items()]
    assert [type(k) == uuid.uuid4 and isinstance(v, vhq.Edge) for k, v in task_graph.edges.items()]
    assert critical_path is not None and duration is not None and paths is not None

    last_task_output, outputs = task_graph.activate()
    assert last_task_output is not None and isinstance(last_task_output, vhq.TaskOutput)
    assert [k in task_graph.nodes.keys() and v and isinstance(v, vhq.TaskOutput) for k, v in outputs.items()]
    assert task_graph.concl == last_task_output and task_graph.concl_response_schema == None
    assert isinstance(task_graph.usage, UsageMetrics)
    assert task_graph.usage.total_tokens > 0



def test_eval():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    task = vhq.Task(
        description="Research a topic to teach a kid aged 6 about math.",
        response_schema=CustomOutput,
        should_evaluate=True, # triggers evaluation
        eval_criteria=["Uniquness", "Fit to audience",],

    )
    res = task.execute()

    assert isinstance(res.evaluation, vhq.Evaluation)
    assert len(res.evaluation.items) == 2
    assert res.evaluation.aggregate_score is not None
    assert res.evaluation.suggestion_summary is not None



"""frontend home use cases pre-test"""

def test_demo_hello_world():
    import versionhq as vhq

    task = vhq.Task(description='hello world')
    res = task.execute()

    assert res.raw


def test_demo_network():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task='draft a promotional email for the given client',
        expected_outcome='email subject and body in string',
        context="use email_subject and email_body as keys in your response."
    )
    assert isinstance(network, vhq.AgentNetwork) and isinstance(network.formation, vhq.Formation)

    res, _ = network.launch()
    assert res.pydantic is not None


def test_demo_agent_customization():
    import versionhq as vhq
    from pathlib import Path

    current_dir = Path(__file__).parent
    file_path = current_dir / "_sample/sample.csv"

    agent = vhq.Agent(role='Demo Manager')
    agent.update(
        llm='gemini-2.0',
        llm_config = dict(
            temperature=1,
            top_p=0.1,
            n=1,
            stop="test",
        ),
        knowledge_sources = [
            'https://business.linkedin.com',
            file_path
        ],
        with_memory=True
    )

    assert "gemini-2.0" in agent.llm.model
    assert agent.llm.llm_config["temperature"] == 1
    assert agent.llm.llm_config["top_p"] == 0.1
    assert agent.llm.llm_config["n"] == 1
    assert agent.llm.llm_config["stop"] == "test"
    assert agent.knowledge_sources == ['https://business.linkedin.com', file_path]
    assert agent.with_memory == True

    res = agent.start()
    assert isinstance(res, vhq.TaskOutput)
    assert res.final is not None


def test_solo_tg_eval():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task="test",
        expected_outcome="test"
    )

    res, tg = network.launch()
    eval = tg.evaluate(
        eval_criteria=["cost", "", "{criteria_3}"]
    )
    assert isinstance(eval, vhq.Evaluation)


def test_llm_as_judge():
    from versionhq._utils.llm_as_a_judge import LLMJudge, generate_summaries, validate
    class MockSummarizer:
        def summarize(self, text: str) -> str:
            return f"Summary of: {text[:50]}..."

    from pathlib import Path
    current_dir = Path(__file__).parent
    file_path = current_dir / "_sample/sample.json"
    summaries = generate_summaries(file_path=file_path, summarizer=MockSummarizer())
    results = validate(judge=LLMJudge(), data=summaries, threshold=0.6)

    assert results is not None


def _test_collab():
    import versionhq as vhq
    from pathlib import Path

    current_dir = Path(__file__).parent
    file_path = current_dir / "_sample/screenshot.png"

    def custom_tool(query: str): return query

    agent = vhq.Agent(
        role="DEMO Computer Use",
        goal="Make a query to check the weather using the web browser. The screenshot of the web browser will be provided.",
        llm="gemini-2.0",
        tools=[custom_tool,],
        maxit=1,
        max_retry_limit=1
    )
    assert [item for item in agent.tools if isinstance(item, vhq.Tool)]

    res = agent.start(image=str(file_path))
    assert res.json_dict is not None



import pytest

from versionhq.task.model import Task
from tests import demo_response_fields


@pytest.fixture(scope='module')
def sub_task():
    return Task(description="write a random poem.")


@pytest.fixture(scope='module')
def main_task():
    from pathlib import Path

    current_dir = Path(__file__).parent.parent
    file_path = current_dir / "_sample/screenshot.png"
    audio_path = current_dir / "_sample/sample.mp3"

    return Task(
        description="return random values strictly following the given response format.",
        response_schema=demo_response_fields,
        tools=[lambda x: x, lambda y: y],
        can_use_agent_tools=True,
        image=str(file_path),
        file=str(file_path),
        audio=str(audio_path),
        should_evaluate=True,
        eval_criteria=["test1", "test2"],
    )


@pytest.fixture(scope='module')
def rag_tool():
    import versionhq as vhq
    rag_tool = vhq.RagTool(url="https://github.com/chroma-core/chroma/issues/3233", query="What is the next action plan?")
    return rag_tool


def test_draft_prompt(main_task, sub_task, rag_tool):
    import versionhq as vhq
    from versionhq._prompt.model import Prompt

    agent = vhq.Agent(llm="gemini-2.0", role="Content Interpretator")
    user_prompt, dev_prompt, messages = Prompt(task=main_task, agent=agent, context=["test", "test2", sub_task]).format_core(rag_tools=[rag_tool])

    assert user_prompt is not None
    assert dev_prompt is not None
    assert messages[0]["role"] == "user"
    assert isinstance(messages[0]["content"], list)
    assert messages[1]["role"] == "developer"
    assert messages[1]["content"] == agent.backstory


def test_draft_prompt_with_tools(main_task, sub_task, rag_tool):
    import versionhq as vhq
    from versionhq._prompt.model import Prompt

    agent = vhq.Agent(llm="gemini-2.0", role="Content Interpretator", tools=[rag_tool])
    user_prompt, dev_prompt, messages = Prompt(task=main_task, agent=agent, context=["test", "test2", sub_task]).format_core()

    assert user_prompt is not None
    assert dev_prompt is not None
    assert messages[0]["role"] == "user"
    assert isinstance(messages[0]["content"], list)
    assert messages[1]["role"] == "developer"
    assert messages[1]["content"] == agent.backstory



from pathlib import Path
from typing import Iterator, List

try:
    from docling.datamodel.base_models import InputFormat
    from docling.document_converter import DocumentConverter
    from docling.exceptions import ConversionError
    from docling_core.transforms.chunker.hierarchical_chunker import HierarchicalChunker
    from docling_core.types.doc.document import DoclingDocument
    DOCLING_AVAILABLE = True
# except ImportError:
    # import envoy
    # envoy.run("uv add docling --optional docling")
    # DOCLING_AVAILABLE = True
except:
    DOCLING_AVAILABLE = False

from pydantic import Field

from versionhq.knowledge.source import BaseKnowledgeSource
from versionhq.storage.utils import fetch_db_storage_path
from versionhq._utils import KNOWLEDGE_DIRECTORY, is_valid_url


class DoclingSource(BaseKnowledgeSource):
    """
    Default docling class for converting documents to markdown or json.
    Support PDF, DOCX, TXT, XLSX, PPTX, MD, Images, and HTML files without any additional dependencies.
    """

    file_paths: List[Path | str] = Field(default_factory=list)
    valid_file_paths: List[Path | str] = Field(default_factory=list)
    content: List["DoclingDocument"] = Field(default_factory=list)
    document_converter: "DocumentConverter" = Field(default_factory=lambda: DocumentConverter(
        allowed_formats=[
            InputFormat.MD,
            InputFormat.ASCIIDOC,
            InputFormat.PDF,
            InputFormat.DOCX,
            InputFormat.HTML,
            InputFormat.IMAGE,
            InputFormat.XLSX,
            InputFormat.PPTX,
        ]
    ))

    def __init__(self, *args, **kwargs):
        if DOCLING_AVAILABLE:
            from docling.datamodel.base_models import InputFormat
            from docling.document_converter import DocumentConverter
            from docling.exceptions import ConversionError
            from docling_core.transforms.chunker.hierarchical_chunker import HierarchicalChunker
            from docling_core.types.doc.document import DoclingDocument

            super().__init__(*args, **kwargs)

        else:
            raise ImportError("The docling package is required. Please install the package using: $ uv add versionhq[docling]")
        # else:
        #     super().__init__(*args, **kwargs)


    def _convert_source_to_docling_documents(self) -> List["DoclingDocument"]:
        conv_results_iter = self.document_converter.convert_all(self.valid_file_paths)
        return [result.document for result in conv_results_iter]


    def _load_content(self) -> List["DoclingDocument"]:
        try:
            return self._convert_source_to_docling_documents()
        except ConversionError as e:
            self._logger.log(level="error", message=f"Error loading content: {str(e)}. Supported formats: {self.document_converter.allowed_formats}", color="red")
            raise e
        except Exception as e:
            self._logger.log(level="error", message=f"Error loading content: {str(e)}", color="red")
            raise e


    def _chunk_doc(self, doc: "DoclingDocument") -> Iterator[str]:
        chunker = HierarchicalChunker()
        for chunk in chunker.chunk(doc):
            yield chunk.text


    def model_post_init(self, _) -> None:
        self.valid_file_paths = self.validate_content()
        self.content.extend(self._load_content())


    def validate_content(self) -> List[Path | str]:
        processed_paths: List[Path | str] = []
        for path in self.file_paths:
            if isinstance(path, str):
                if path.startswith(("http://", "https://")):
                    try:
                        if is_valid_url(path):
                            processed_paths.append(path)
                        else:
                            raise ValueError(f"Invalid URL format: {path}")
                    except Exception as e:
                        raise ValueError(f"Invalid URL: {path}. Error: {str(e)}")
                else:
                    local_path = Path(KNOWLEDGE_DIRECTORY + "/" + path)
                    if local_path.exists():
                        processed_paths.append(local_path)
                    else:
                        local_path = Path(fetch_db_storage_path() + "/" + KNOWLEDGE_DIRECTORY + "/" + path) # try with abs. path
                        if local_path.exists():
                            processed_paths.append(local_path)
                        else:
                            raise FileNotFoundError(f"File not found: {local_path}")
            else:
                if isinstance(path, Path):
                    processed_paths.append(path)
        return processed_paths


    def add(self) -> None:
        if self.content is None:
            self.model_post_init()

        if self.content:
            for doc in self.content:
                new_chunks_iterable = self._chunk_doc(doc)
                self.chunks.extend(list(new_chunks_iterable))
            self._save_documents()



import uuid
import warnings
from enum import IntEnum
from hashlib import md5
from typing import Any, Dict, List, Callable, Optional, Tuple
from typing_extensions import Self

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator
from pydantic._internal._generate_schema import GenerateSchema
from pydantic_core import PydanticCustomError, core_schema


from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput, TaskExecutionType, ResponseField
from versionhq.task_graph.model import TaskGraph, Node, Edge, TaskStatus, DependencyType, Condition
from versionhq._utils.logger import Logger
# from versionhq.recording.usage_metrics import UsageMetrics


initial_match_type = GenerateSchema.match_type

def match_type(self, obj):
    if getattr(obj, "__name__", None) == "datetime":
        return core_schema.datetime_schema()
    return initial_match_type(self, obj)


GenerateSchema.match_type = match_type
warnings.filterwarnings("ignore", category=SyntaxWarning, module="pysbd")


class Formation(IntEnum):
    SOLO = 1
    SUPERVISING = 2
    SQUAD = 3
    RANDOM = 4
    HYBRID = 10


class TaskHandlingProcess(IntEnum):
    """
    A class representing task handling processes to tackle multiple tasks.
    When the agent network has multiple tasks that connect with edges, follow the edge conditions.
    """
    HIERARCHY = 1
    SEQUENTIAL = 2
    CONSENSUAL = 3 # either from manager class agents from diff network or human. need to define a trigger


class Member(BaseModel):
    """
    A class to store a member (agent) in the network, with its tasks and memory/knowledge share settings.
    """
    agent: Agent | None = Field(default=None)
    is_manager: bool = Field(default=False)
    can_share_knowledge: bool = Field(default=True, description="whether to share the agent's knowledge in the network")
    can_share_memory: bool = Field(default=True, description="whether to share the agent's memory in the network")
    tasks: Optional[List[Task]] = Field(default_factory=list, description="tasks explicitly assigned to the agent")

    @property
    def is_idling(self):
        return bool(self.tasks)


class AgentNetwork(BaseModel):
    """A Pydantic class to store an agent network with agent members and tasks."""

    __hash__ = object.__hash__
    _execution_span: Any = PrivateAttr()
    _inputs: Optional[Dict[str, Any]] = PrivateAttr(default=None)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None)
    members: List[Member] = Field(default_factory=list)
    formation: Optional[Formation] = Field(default=None)
    should_reform: bool = Field(default=False, description="True if task exe. failed or eval scores below threshold")

    network_tasks: Optional[List[Task]] = Field(default_factory=list, description="tasks without dedicated agents - network's common tasks")

    # task execution rules
    prompt_file: str = Field(default="", description="absolute file path to the prompt file that stores jsonified prompts")
    process: TaskHandlingProcess = Field(default=TaskHandlingProcess.SEQUENTIAL)
    consent_trigger: Optional[Callable | Condition] = Field(default=None, description="returns bool")

    # callbacks
    pre_launch_callbacks: List[Callable[..., Any]]= Field(default_factory=list, description="list of callback funcs called before the network launch")
    post_launch_callbacks: List[Callable[..., Any]] = Field(default_factory=list, description="list of callback funcs called after the network launch")
    step_callback: Optional[Any] = Field(default=None, description="callback to be executed after each step of all agents execution")

    cache: bool = Field(default=True)
    execution_logs: List[Dict[str, Any]] = Field(default_factory=list, description="list of execution logs of the tasks handled by members")


    def __name__(self) -> str:
        return self.name if self.name is not None else self.id.__str__


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        """Prevent manual setting of the 'id' field by users."""
        if v:
            raise PydanticCustomError("may_not_set_field", "The 'id' field cannot be set by the user.", {})


    @model_validator(mode="after")
    def validate_process(self) -> Self:
        if self.process == TaskHandlingProcess.CONSENSUAL and not self.consent_trigger:
            Logger().log(level="error", message="Need to define the consent trigger function that returns bool", color="red")
            raise PydanticCustomError("invalid_process", "Need to define the consent trigger function that returns bool", {})


        if self.consent_trigger and isinstance(self.consent_trigger, Callable):
            self.consent_trigger = Condition(methods={"0": self.consent_trigger})
        return self


    @model_validator(mode="after")
    def set_up_network_for_members(self) -> Self:
        if self.members:
            for member in self.members:
                member.agent.networks.append(self)

        return self


    @model_validator(mode="after")
    def assess_tasks(self):
        """
        Validates if the model recognize all tasks that the network needs to handle.
        """

        if self.tasks:
            if all(task in self.tasks for task in self.network_tasks) == False:
                raise PydanticCustomError("task_validation_error", "`network_tasks` needs to be recognized in the task.", {})


            num_member_tasks = 0
            for member in self.members:
                num_member_tasks += len(member.tasks)

            # if len(self.tasks) != len(self.network_tasks) + num_member_tasks:
            #     raise PydanticCustomError("task_validation_error", "Some tasks are missing.", {})
        return self


    @model_validator(mode="after")
    def check_manager(self):
        """
        Check if the agent network has a manager
        """
        if self.process == TaskHandlingProcess.HIERARCHY or self.formation == Formation.SUPERVISING:
            if not self.managers:
                Logger().log(level="error", message="The process or formation created needs at least 1 manager agent.", color="red")
                raise PydanticCustomError("missing_manager", "`manager` is required when using hierarchical process.", {})

        ## comment out for the formation flexibilities
        # if self.managers and (self.manager_tasks is None or self.network_tasks is None):
        #     Logger().log(level="error", message="The manager is idling. At least 1 task needs to be assigned to the manager.", color="red")
        #     raise PydanticCustomError("missing_manager_task", "manager needs to have at least one manager task or network task.", {})

        return self


    @model_validator(mode="after")
    def validate_task_member_paring(self):
        """
        Sequential task processing without any network_tasks require a task-agent pairing.
        """
        if self.process == TaskHandlingProcess.SEQUENTIAL and self.network_tasks is None:
            for task in self.tasks:
                if not [member.task == task for member in self.members]:
                    Logger().log(level="error", message=f"The following task needs a dedicated agent to be assinged: {task.description}", color="red")
                    raise PydanticCustomError("missing_agent_in_task", "Sequential process error: Agent is missing the task", {})
        return self


    @model_validator(mode="after")
    def validate_end_with_at_most_one_async_task(self):
        """
        Validates that the agent network completes max. one asynchronous task by counting tasks traversed backward
        """

        async_task_count = 0
        for task in reversed(self.tasks):
            if not task:
                break
            elif task.execution_type == TaskExecutionType.ASYNC:
                async_task_count += 1
            else:
                break

        if async_task_count > 1:
            raise PydanticCustomError("async_task_count", "The agent network must end with at maximum one asynchronous task.", {})
        return self


    def _generate_agents(self, unassigned_tasks: List[Task]) -> List[Member]:
        """Generates agents as members in the network."""

        from versionhq.agent.inhouse_agents import vhq_agent_creator

        new_member_list: List[Member] = []

        for unassgined_task in unassigned_tasks:
            task = Task(
                description=f"Based on the following task summary, draft an agent's role and goal in concise manner. Task summary: {unassgined_task.summary}",
                response_schema=[
                    ResponseField(title="goal", data_type=str, required=True),
                    ResponseField(title="role", data_type=str, required=True),
                ],
            )
            res = task.execute(agent=vhq_agent_creator)
            agent = Agent(
                role=res.json_dict["role"] if "role" in res.json_dict else res.raw,
                goal=res.json_dict["goal"] if "goal" in res.json_dict else task.description
            )
            if agent.id:
                member = Member(agent=agent, tasks=[unassgined_task], is_manager=False)
                new_member_list.append(member)

        return new_member_list


    def _assign_tasks(self) -> None:
        """Assigns tasks to member agents."""

        idling_managers: List[Member] = [member for member in self.members if member.is_idling and member.is_manager == True]
        idling_members: List[Member] =  [member for member in self.members if member.is_idling and member.is_manager == False]
        unassigned_tasks: List[Task] = self.network_tasks + self.unassigned_member_tasks if self.network_tasks else self.unassigned_member_tasks
        new_members: List[Member] = []

        if not unassigned_tasks:
            return

        else:
            if idling_managers:
                idling_managers[0].tasks.extend(unassigned_tasks)

            elif idling_members:
                idling_members[0].tasks.extend(unassigned_tasks)

            else:
                new_members = self._generate_agents(unassigned_tasks=unassigned_tasks)
                if new_members:
                    self.members += new_members


    def _get_responsible_agent(self, task: Task) -> Agent | None:
        if not task:
            return None

        self._assign_tasks()
        for member in self.members:
            if member.tasks and [item for item in member.tasks if item.id == task.id]:
                return member.agent


    def _execute_tasks(self, tasks: List[Task], start_index: Optional[int] = None) -> Tuple[TaskOutput, TaskGraph]:
        """Executes tasks and returns TaskOutput object as concl or latest response in the network."""
        res, task_graph = None, None

        if len(tasks) == 1:
            task = self.tasks[0]
            responsible_agent = self._get_responsible_agent(task=task)
            res = task.execute(agent=responsible_agent)
            node = Node(task=task)
            task_graph = TaskGraph(nodes={ node.identifier: node, }, concl=res if res else None)
            return res, task_graph

        nodes = [
            Node(
                task=task,
                assigned_to=self._get_responsible_agent(task=task),
                status=TaskStatus.NOT_STARTED if not start_index or i >= start_index else TaskStatus.COMPLETED,
            ) for i, task in enumerate(tasks)
        ]
        task_graph = TaskGraph(nodes={node.identifier: node for node in nodes})

        for i in range(0, len(nodes) - 1):
            condition = self.consent_trigger if isinstance(self.consent_trigger, Condition) else Condition(methods={"0": self.consent_trigger }) if self.consent_trigger else None
            task_graph.add_edge(
                source=nodes[i].identifier,
                target=nodes[i+1].identifier,
                edge=Edge(
                    weight=3 if nodes[i].task in self.manager_tasks else 1,
                    dependency_type=DependencyType.FINISH_TO_START if self.process == TaskHandlingProcess.HIERARCHY else DependencyType.START_TO_START,
                    required=bool(self.process == TaskHandlingProcess.CONSENSUAL),
                    condition=condition,
                    data_transfer=bool(self.process == TaskHandlingProcess.HIERARCHY),
                )
            )

        if start_index is not None:
            res, _ = task_graph.activate(target=nodes[start_index].indentifier)

        else:
            res, _ = task_graph.activate()

        if not res:
            Logger().log(level="error", message="Missing task outputs.", color="red")
            raise ValueError("Missing task outputs")

        return res, task_graph


    def launch(self, kwargs_pre: Optional[Dict[str, str]] = None, kwargs_post: Optional[Dict[str, Any]] = None, start_index: int = None) -> Tuple[TaskOutput, TaskGraph]:
        """Launches agent network by executing tasks in the network and recording the outputs"""

        self._assign_tasks()

        if kwargs_pre:
            for func in self.pre_launch_callbacks: #! REFINEME -  signature check
                func(**kwargs_pre)

        for member in self.members:
            agent = member.agent

            if not agent.networks:
                agent.networks.append(self)

            if self.step_callback:
                agent.callbacks.append(self.step_callback)

        if self.process is None:
            self.process = TaskHandlingProcess.SEQUENTIAL

        result, tg = self._execute_tasks(self.tasks, start_index=start_index)
        callback_output = None

        for func in self.post_launch_callbacks:
            callback_output = func(result, **kwargs_post)

        if callback_output:
            match result:
                case TaskOutput():
                    result.callback_output = callback_output

                case _:
                    pass

        return result, tg


    @property
    def key(self) -> str:
        source = [str(member.agent.id.__str__) for member in self.members] + [str(task.id.__str__) for task in self.tasks]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def managers(self) -> List[Member] | None:
        managers = [member for member in self.members if member.is_manager == True]
        return managers if len(managers) > 0 else None


    @property
    def manager_tasks(self) -> List[Task]:
        """
        Tasks (incl. network tasks) handled by managers in the agent network.
        """
        res = list()

        if self.managers:
            for manager in self.managers:
                if manager.tasks:
                    res.extend(manager.tasks)

        return res


    @property
    def tasks(self) -> List[Task]:
        """
        Return all the tasks that the agent network needs to handle in order of priority:
        1. network_tasks, -> assigned to the member
        2. manager_task,
        3. members' tasks
        """

        network_tasks = self.network_tasks
        manager_tasks = self.manager_tasks
        member_tasks = []

        for member in self.members:
            if member.is_manager == False and member.tasks:
                a = [item for item in member.tasks if item not in network_tasks and item not in manager_tasks]
                member_tasks += a

        return network_tasks + manager_tasks + member_tasks


    @property
    def unassigned_member_tasks(self) -> List[Task]:
        res = list()

        if self.members:
            for member in self.members:
                if member.agent is None and member.tasks:
                    res.extend(member.tasks)

        return res



import uuid
from abc import ABC
from typing import Optional, List

from pydantic import UUID4, BaseModel, Field, field_validator, model_validator
from pydantic_core import PydanticCustomError

from versionhq.tool.composio.params import ComposioAppName


class ProductProvider(ABC, BaseModel):
    """
    Abstract class for the product provider entity.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None)
    region: Optional[str] = Field(default=None, description="region of client's main business operation")
    data_pipelines: Optional[List[ComposioAppName | str]] = Field(default=None)
    destination_services: Optional[List[ComposioAppName | str]] = Field(default=None)

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def set_up_destinations(self):
        """
        Set up the destination services and data pipeines using ComposioAppName class.
        """
        if self.destination_services is not None:
            results = []
            for item in self.destination_services:
                if isinstance(item, ComposioAppName):
                    results.append(item)
                elif isinstance(item, str) and item in ComposioAppName:
                    results.append(ComposioAppName(item))
                else:
                    results.append(item)
            self.destination_services = results

        if self.data_pipelines is not None:
            results = []
            for item in self.data_pipelines:
                if isinstance(item, ComposioAppName):
                    results.append(item)
                elif isinstance(item, str) and item in ComposioAppName:
                    results.append(ComposioAppName(item))
                else:
                    results.append(item)
            self.data_pipelines = results

        return self


class Product(BaseModel):
    """
    A class to store product information used to create outbound
    """
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="product name")
    description: Optional[str] = Field(
        default=None,
        max_length=256,
        description="product description scraped from landing url or client input. cascade to the agent"
    )
    provider: Optional[ProductProvider] = Field(default=None)
    audience: Optional[str] = Field(default=None, description="target audience")
    usp: Optional[str] = Field(default=None)
    landing_url: Optional[str] = Field(default=None, description="marketing url of the product if any")
    cohort_timeframe: Optional[int] = Field(default=30, description="ideal cohort timeframe of the product in days")
    notes: Optional[str] = Field(default=None, description="any notes from the client to consider. cascade to the agent")


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @field_validator("cohort_timeframe", mode="before")
    @classmethod
    def _deny_non_int_input(cls, v: Optional[UUID4]) -> None:
        if not isinstance(v, int):
            raise PydanticCustomError("invalid_input", "This field only accepts inputs in integer.", {})
